@@ -22,6 +22,8 @@
 #ifndef COLLECTION_H
 #define COLLECTION_H
 
+#include <stdint.h>
+
 #ifndef EOK
 #define EOK 0
 #endif
@@ -49,10 +51,6 @@
                                                collections */
 
 
-/* Any data we deal with can't be longer than this */
-/* FIXME - make it compile time option */
-#define COL_MAX_DATA    65535
-
 /* Default class for a free form collection */
 #define COL_CLASS_DEFAULT      0
 
@@ -70,10 +68,18 @@
                                      * fail. */
 #define COL_ADD_MODE_CLONE     2    /* Creates a deep copy of a collection with
                                      * its sub collections */
-
 #define COL_ADD_MODE_FLAT      3    /* Creates a deep copy of a collection with
-                                     * its sub collections flattening and resolving
-                                     * duplicates.
+                                     * its sub collections flattening and NOT
+                                     * resolving duplicates.
+                                     */
+#define COL_ADD_MODE_FLATDOT   4    /* Creates a deep copy of a collection with
+                                     * its sub collections flattening and NOT
+                                     * resolving duplicates. Names are contructed
+                                     * in dotted notation.
+                                     * For example the subcollection
+                                     * named "sub" containing "foo" and
+                                     * "bar" will be flattened as:
+                                     * "sub.foo", "sub.bar".
                                      */
 
 /* Modes how the collection is traversed */
@@ -90,6 +96,7 @@
                                            * is present */
 #define COL_TRAVERSE_FLAT     0x00000008  /* Flatten the collection. */
 
+
 /* Additional iterator flags
  * NOTE: ignored by traverse functions */
 #define COL_TRAVERSE_SHOWSUB  0x00010000  /* Include headers of sub collections
@@ -112,11 +119,21 @@
  * DO NOT MIX THEM IN ONE ITERATOR.
  */
 
-/* FIXME: move to event level - this does not belong to collection */
-/* Time stamp property name */
-#define TS_NAME         "stamp"
-/* Time property name */
-#define T_NAME          "time"
+
+/* Modes accepted by copy collection function */
+#define COL_COPY_NORMAL         0    /* Deep copy. Referenced collections
+                                      * of the donor are copied as sub
+                                      * collections.
+                                      */
+#define COL_COPY_FLAT           1    /* Deep copy. Collection is flattend. */
+#define COL_COPY_FLATDOT        2    /* Deep copy. Collection is flattend.
+                                      * Names are concatenated with dot.
+                                      */
+#define COL_COPY_KEEPREF        3    /* Deep copy but leave references
+                                      * as references.
+                                      */
+#define COL_COPY_TOP            4    /* Copy only top level
+                                      */
 
 /* Match values */
 #define COL_NOMATCH 0
@@ -198,21 +215,6 @@ int col_create_collection(struct collection_item **ci,
                           const char *name,
                           unsigned cclass);
 
-/* Function that creates a named collection using a memory descriptor */
-/* FIXME - function is a placeholder. It is not implemented yet.
- * will be added in future together with the definition of the
- * descriptor structure.
- * The purpose is to control the internal implementation of the collection
- * a) Use hash table for faster searches if the collection is expected to be large.
- * b) Define memory functions to use.
- */
-/*
-int col_create_collection_ex(struct collection_item **ci,
-                             const char *name,
-                             unsigned cclass,
-                             struct cdescriptor *descrptor);
-*/
-
 /* Function that destroys a collection */
 void col_destroy_collection(struct collection_item *ci);
 
@@ -315,15 +317,6 @@ int col_add_any_property_with_ref(struct collection_item *ci,
                                   int type, void *data, int length,
                                   struct collection_item **ret_ref);
 
-/* FIXME - does not belong here - move to other place */
-/* Function to create a timestamp */
-/* Automatically adds/updates time and timestamp properties in the
- * collection returning references */
-int col_set_timestamp(struct collection_item *ci,
-                      struct collection_item **timestr_ref,
-                      struct collection_item **timeint_ref);
-
-
 /* Update functions */
 /* All update functions search the property using the search algorithm
  * described at the top of the header file.
@@ -397,11 +390,31 @@ int col_add_collection_to_collection(struct collection_item *ci,            /* C
                                  struct collection_item *collection_to_add, /* Collection to add */
                                  int mode);                                 /* How this collection needs to be added */
 
-/* Create a deep copy of the current collection. */
-/* Referenced collections of the donor are copied as sub collections. */
+/* Create a deep copy of the current collection.
+ * Wraps the function below.
+ * The acceptable modes are defined at the top.
+ */
 int col_copy_collection(struct collection_item **collection_copy,
                         struct collection_item *collection_to_copy,
-                        const char *name_to_use);
+                        const char *name_to_use,
+                        int copy_mode);
+
+/* Callback used in the next function */
+typedef int (*col_copy_cb)(struct collection_item *item,
+                           void *ext_data,
+                           int *skip);
+
+/* Create a deep copy of the current collection.
+ * Calls caller provided callback before
+ * copying each item's data.
+ * The acceptable modes are defined at the top.
+ */
+int col_copy_collection_with_cb(struct collection_item **collection_copy,
+                                struct collection_item *collection_to_copy,
+                                const char *name_to_use,
+                                int copy_mode,
+                                col_copy_cb copy_cb,
+                                void *ext_data);
 
 /* Signature of the callback that needs to be used when
    traversing a collection or looking for a specific item */
@@ -460,6 +473,125 @@ const char *col_get_item_property(struct collection_item *ci, int *property_len)
 int col_get_item_type(struct collection_item *ci);
 int col_get_item_length(struct collection_item *ci);
 void *col_get_item_data(struct collection_item *ci);
+uint64_t col_get_item_hash(struct collection_item *ci);
+
+/* Calculates hash of the string using internal hashing
+ * algorithm. Populates "length" with length
+ * of the string not counting 0.
+ * Length argument can be NULL.
+ * If sub_len is greater than zero
+ * this value is used to count how many characters
+ * from string should be included into hash
+ * calculation.
+ */
+uint64_t col_make_hash(const char *string, int sub_len, int *length);
+
+/* Compare two items.
+ * The second item is evaluated against the first.
+ * Function returns 0 if two items are the same
+ * and non-zero otherwise.
+ * The in_flags is a bit mask that defines
+ * how the items should be compared.
+ * See below the list of conbstants
+ * defined for this purpose.
+ * If items are different they might be orderable
+ * or not. For example one can order items by name
+ * but not by type.
+ * If the result of the function is non-zero
+ * the out_flags (if provided) will be
+ * set to indicate if the second item is greater
+ * then the first.
+ */
+int col_compare_items(struct collection_item *first,
+                      struct collection_item *second,
+                      unsigned in_flags,
+                      unsigned *out_flags);
+
+/********* Possible valies for input flags ********/
+/* How to compare properties?
+ * The following 4 flags are mutually exclusive
+ */
+#define COL_CMPIN_PROP_EQU    0x000000004 /* Properties should be same */
+#define COL_CMPIN_PROP_BEG    0x000000005 /* Properties start same */
+#define COL_CMPIN_PROP_MID    0x000000006 /* One is substring of another */
+#define COL_CMPIN_PROP_END    0x000000007 /* One property ends with another */
+
+/* Make sure that there is a dot.
+ * Useful with _BEG, _MID and _END flags to check that the there is
+ * a dot (if present) in the right place (before, after or both).
+ * For example the first item is named "foo.bar" and the second
+ * is "bar". Using _END the "bar" will be found but if _DOT flag is
+ * used too the function will also check if there was a "." before the found
+ * string in this case.
+ * Ignored in case of _EQU.
+ */
+#define COL_CMPIN_PROP_DOT     0x000000008
+
+/* Compare property lenghts */
+#define COL_CMPIN_PROP_LEN     0x000000010
+
+/* Compare types */
+#define COL_CMPIN_TYPE         0x000000020
+
+/* Compare data len */
+#define COL_CMPIN_DATA_LEN     0x000000040
+
+/* Compare data (up to the length of the second one)
+ * if type is the same. If type is different
+ * function will assume data is different
+ * without performing actual comparison.
+ */
+#define COL_CMPIN_DATA         0x000000080
+
+/********* Possible values for output flags *********/
+/* If _EQU was specified and the property of the second item
+ * is greater the following bit will be set
+ */
+#define COL_CMPOUT_PROP_STR    0x00000001
+/* If we were told to compare property lengths
+ * and second is longer this bit will be set
+ */
+#define COL_CMPOUT_PROP_LEN    0x00000002
+/* If we were told to compare data lengths
+ * and second is longer this bit will be set
+ */
+#define COL_CMPOUT_DATA_LEN    0x00000004
+/* If we were told to compare data
+ * and types are the same then
+ * if the second one is greater this bit will
+ * be set. If data is binary flag is never set
+ */
+#define COL_CMPOUT_DATA    0x00000008
+
+
+/* Sort collection.
+ * cmp_flags are the same as in_flags for the compare
+ * function. The sort_flags is an OR of the costants
+ * defined below.
+ * If the subcollections are included in sorting
+ * each collection is sorted separately (this is not a global sort).
+ * It might be dangerous to sort subcollections if
+ * subcollection is not owned by current collection.
+ * If it is a reference to an external collection
+ * there might be an issue. To skip the collections that
+ * externally referenced use MYSUB flag.
+ * Keep in mind that if the collection
+ * has two references to the same other
+ * collection it is impossible to detect
+ * this situation. If MYSUB is used in this
+ * case such collection will be ignored
+ * If MYSUB is not used the collection
+ * will be sorted more than once.
+ */
+int col_sort_collection(struct collection_item *col,
+                        unsigned cmp_flags,
+                        unsigned sort_flags);
+
+/* Sort flags */
+#define COL_SORT_ASC    0x00000000
+#define COL_SORT_DESC   0x00000001
+#define COL_SORT_SUB    0x00000002
+#define COL_SORT_MYSUB  0x00000004
 
 /* If you want to modify the item that you got as a result of iterating through collection
  * or by calling col_get_item(). If you want to rename item provide a new name in the property
@@ -535,8 +667,10 @@ int col_is_item_in_collection(struct collection_item *ci,   /* Collection to fin
                               int *found);                  /* Boolean that turns to nonzero if the match is found */
 
 
-/* Get collection - get a pointer to a collection included into another collection */
-/* Delete extracted collection after use to decrease reference count. */
+/* Get collection - get a pointer to a collection included into another collection.
+ * If the collection_to_find is NULL function reterns a reference to the top level collection.
+ * Delete extracted collection after use to decrease reference count.
+ */
 int col_get_collection_reference(struct collection_item *ci,          /* High level collection */
                                  struct collection_item **acceptor,   /* The pointer that will accept extracted handle */
                                  const char *collection_to_find);     /* Name to of the collection */
@@ -595,7 +729,11 @@ int col_get_iterator_depth(struct collection_iterator *iterator, int *depth);
  */
 int col_get_item_depth(struct collection_iterator *iterator, int *depth);
 
-/* FIXME - Do we need to be able to rewind iterator? */
+/* Pins down the iterator to loop around current point */
+void col_pin_iterator(struct collection_iterator *iterator);
+
+/* Rewinds iterator to the beginning */
+void col_rewind_iterator(struct collection_iterator *iterator);
 
 /* Set collection class */
 int col_set_collection_class(struct collection_item *item, /* Collection */
