@@ -117,7 +117,7 @@ int col_debug_handle(const char *property,
                (nest_level -1) * 4, "",
                property,
                length,
-               (*((unsigned char *)(data)) == '\0') ? "flase" : "true",
+               (*((unsigned char *)(data)) == '\0') ? "false" : "true",
                nest_level);
         break;
     case COL_TYPE_COLLECTION:
@@ -244,7 +244,7 @@ int col_get_data_len(int type, int length)
     case COL_TYPE_UNSIGNED:
     case COL_TYPE_LONG:
     case COL_TYPE_ULONG:
-        len = 15;
+        len = 20;
         break;
 
     case COL_TYPE_STRING:
