@@ -33,6 +33,7 @@ int ref_collection_test(void)
 {
     struct collection_item *peer = NULL;
     struct collection_item *socket = NULL;
+    struct collection_item *socket2 = NULL;
     char binary_dump[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8 };
 
     int error = EOK;
@@ -98,7 +99,19 @@ int ref_collection_test(void)
     col_destroy_collection(peer);
 
     col_debug_collection(socket, COL_TRAVERSE_DEFAULT);
+
+    error = col_get_collection_reference(socket, &socket2, NULL);
+    if (error) {
+        col_destroy_collection(socket);
+        printf("Failed to extract collection. Error %d\n", error);
+        return error;
+    }
+
+    col_debug_collection(socket2, COL_TRAVERSE_DEFAULT);
     col_destroy_collection(socket);
+    col_debug_collection(socket2, COL_TRAVERSE_DEFAULT);
+    col_destroy_collection(socket2);
+
     TRACE_FLOW_NUMBER("ref_collection_test. Returning", error);
 
     printf("\n\nEND OF REF TEST!!!.\n\n\n");
@@ -222,16 +235,30 @@ int add_collection_test(void)
     return error;
 }
 
+int copy_cb(struct collection_item *item,
+            void *ext_data,
+            int *skip)
+{
+    printf("INSIDE Copy Callback\n");
+    col_debug_item(item);
+    printf("Passed in data: %s\n", (char *) ext_data);
+    if (strcmp(col_get_item_property(item, NULL), "id") == 0) *skip = 1;
+    return EOK;
+}
+
+
 int mixed_collection_test(void)
 {
     struct collection_item *peer;
     struct collection_item *socket1;
     struct collection_item *socket2;
+    struct collection_item *socket3;
     struct collection_item *event;
     struct collection_item *host;
     char binary_dump[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8 };
     int found = 0;
     unsigned int class = 0;
+    char foo[] = "foo";
 
     int error = EOK;
 
@@ -283,7 +310,7 @@ int mixed_collection_test(void)
     col_debug_collection(socket1, COL_TRAVERSE_DEFAULT);
     printf("Creating a copy of SOCKET1 collection named SOCKET2.\n");
 
-    error = col_copy_collection(&socket2, socket1, "socket2");
+    error = col_copy_collection(&socket2, socket1, "socket2", COL_COPY_NORMAL);
     if (error) {
         col_destroy_collection(peer);
         col_destroy_collection(host);
@@ -295,6 +322,20 @@ int mixed_collection_test(void)
     col_debug_collection(socket2, COL_TRAVERSE_DEFAULT);
     col_debug_collection(peer, COL_TRAVERSE_DEFAULT);
 
+    error = col_copy_collection_with_cb(&socket3, socket1, "socket3",
+                                        COL_COPY_FLATDOT, copy_cb, (void *)foo);
+    if (error) {
+        col_destroy_collection(peer);
+        col_destroy_collection(host);
+        col_destroy_collection(socket1);
+        col_destroy_collection(socket2);
+        printf("Failed to copy collection. Error %d\n", error);
+        return error;
+    }
+
+    col_debug_collection(socket3, COL_TRAVERSE_DEFAULT);
+    col_destroy_collection(socket3);
+
     printf("Adding PEER collection to SOCKET2 collection as a reference named PEER2\n");
 
     /* Embed peer host into the socket2 as reference */
@@ -564,6 +605,10 @@ int iterator_test(void)
     char binary_dump[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8 };
     int depth = 0;
     int idepth = 0;
+    int len = 0;
+    int i;
+    uint64_t hash1, hash2;
+    int rwnd = 0;
 
     printf("\n\n==== ITERATOR TEST ====\n\n");
 
@@ -640,7 +685,7 @@ int iterator_test(void)
         return error;
     }
 
-    error = col_add_collection_to_collection(peer, NULL, "forth", socket1, COL_ADD_MODE_EMBED);
+    error = col_add_collection_to_collection(peer, NULL, "forth", socket1, COL_ADD_MODE_FLATDOT);
     if (error) {
         col_destroy_collection(peer);
         col_destroy_collection(socket1);
@@ -648,6 +693,16 @@ int iterator_test(void)
         return error;
     }
 
+    error = col_add_collection_to_collection(peer, NULL, NULL, socket1, COL_ADD_MODE_FLATDOT);
+    if (error) {
+        col_destroy_collection(peer);
+        col_destroy_collection(socket1);
+        printf("Failed to add collection to collection. Error %d\n", error);
+        return error;
+    }
+
+    col_destroy_collection(socket1);
+
     /* Bind iterator */
     error =  col_bind_iterator(&iterator, peer, COL_TRAVERSE_DEFAULT);
     if (error) {
@@ -729,6 +784,50 @@ int iterator_test(void)
                 col_destroy_collection(peer);
                 return error;
             }
+
+            printf("Item name: %s\n", col_get_item_property(item, NULL));
+            printf("Item hash: %lu\n", (unsigned long int)col_get_item_hash(item));
+            error = col_modify_item_property(item, "new_name");
+            if (error) {
+                printf("We expected success but got error %d\n", error);
+                col_unbind_iterator(iterator);
+                col_destroy_collection(peer);
+                return error;
+            }
+            len = 0;
+            printf("Item name: %s\n", col_get_item_property(item, &len));
+            printf("Item hash: %lu\n", (unsigned long int)col_get_item_hash(item));
+            printf("Item length: %d\n", len);
+
+            len = 0;
+            hash1 = col_make_hash("new_name", 0, &len);
+            printf("String name: %s\n", "new_name");
+            printf("String hash: %lu\n", (unsigned long int)hash1);
+            printf("String length: %d\n", len);
+
+            len = 0;
+            hash2 = col_make_hash("new_name_suffix", 8, &len);
+            printf("String name: %.*s\n", len, "new_name_suffix");
+            printf("String hash: %lu\n", (unsigned long int)hash2);
+            printf("String length: %d\n", len);
+            if (hash1 != hash2) {
+                printf("Hash calculation failed\n");
+                col_unbind_iterator(iterator);
+                col_destroy_collection(peer);
+                return EINVAL;
+            }
+
+            hash2 = col_make_hash("new_name", 8, &len);
+            printf("String name: %.*s\n", len, "new_name");
+            printf("String hash: %lu\n", (unsigned long int)hash2);
+            printf("String length: %d\n", len);
+            if (hash1 != hash2) {
+                printf("Hash calculation failed\n");
+                col_unbind_iterator(iterator);
+                col_destroy_collection(peer);
+                return EINVAL;
+            }
+
         }
     }
     while(1);
@@ -764,6 +863,13 @@ int iterator_test(void)
         printf("%*s", depth * 4, "");
         col_debug_item(item);
 
+        if ((strcmp(col_get_item_property(item, NULL), "queue") == 0) &&
+            (rwnd == 0)) {
+            printf("Rewinding iterator...\n");
+            col_rewind_iterator(iterator);
+            rwnd++;
+        }
+
     }
     while(1);
 
@@ -956,8 +1062,6 @@ int iterator_test(void)
         return error;
     }
 
-    col_destroy_collection(peer);
-
     printf("\n\nIterate up test:\n\n");
 
     do {
@@ -967,6 +1071,7 @@ int iterator_test(void)
         if (error) {
             printf("Error (iterate): %d\n", error);
             col_unbind_iterator(iterator);
+            col_destroy_collection(peer);
             return error;
         }
 
@@ -1006,6 +1111,174 @@ int iterator_test(void)
     while(1);
 
     col_unbind_iterator(iterator);
+
+    /* Bind iterator again in flat mode */
+    error =  col_bind_iterator(&iterator, peer, COL_TRAVERSE_FLAT | COL_TRAVERSE_END);
+    if (error) {
+        printf("Error (bind): %d\n", error);
+        col_destroy_collection(peer);
+        return error;
+    }
+
+    printf("\n\nCircled looping:\n\n");
+
+    for (i = 0; i < 200; i++) {
+        /* Loop through a collection */
+        error = col_iterate_collection(iterator, &item);
+        if (error) {
+            printf("Error (iterate): %d\n", error);
+            col_destroy_collection(peer);
+            col_unbind_iterator(iterator);
+            return error;
+        }
+
+        /* Are we done ? */
+        if (item == NULL) printf("Reached end.\n\n");
+        else {
+            depth = 0;
+            col_get_item_depth(iterator, &depth);
+            printf("%*s", depth * 4, "");
+            col_debug_item(item);
+        }
+    }
+
+    /* Do not forget to unbind iterator - otherwise there will be a leak */
+    col_unbind_iterator(iterator);
+
+    /* Bind iterator again in flat mode */
+    error =  col_bind_iterator(&iterator, peer, COL_TRAVERSE_FLAT | COL_TRAVERSE_END);
+    if (error) {
+        printf("Error (bind): %d\n", error);
+        col_destroy_collection(peer);
+        return error;
+    }
+
+    printf("\n\nCircled looping with pin:\n\n");
+
+    do {
+        /* Loop through a collection */
+        error = col_iterate_collection(iterator, &item);
+        if (error) {
+            printf("Error (iterate): %d\n", error);
+            col_destroy_collection(peer);
+            col_unbind_iterator(iterator);
+            return error;
+        }
+
+        if (strcmp(col_get_item_property(item, NULL), "queue") == 0) {
+            /* Make it a new looping point */
+            col_pin_iterator(iterator);
+            printf("Found pin point.\n\n");
+            break;
+        }
+        /* Are we done ? */
+        if (item == NULL) {
+            printf("Unexpected end.\n\n");
+            col_destroy_collection(peer);
+            col_unbind_iterator(iterator);
+            return EINVAL;
+        }
+        else {
+            depth = 0;
+            col_get_item_depth(iterator, &depth);
+            printf("%*s", depth * 4, "");
+            col_debug_item(item);
+        }
+    }
+    while(1);
+
+    /* Second loop around the pin point */
+    for (i = 0; i < 200; i++) {
+        /* Loop through a collection */
+        error = col_iterate_collection(iterator, &item);
+        if (error) {
+            printf("Error (iterate): %d\n", error);
+            col_destroy_collection(peer);
+            col_unbind_iterator(iterator);
+            return error;
+        }
+
+        /* Are we done ? */
+        if (item == NULL) printf("Reached end.\n\n");
+        else {
+            depth = 0;
+            col_get_item_depth(iterator, &depth);
+            printf("%*s", depth * 4, "");
+            col_debug_item(item);
+        }
+    }
+
+    /* Do not forget to unbind iterator - otherwise there will be a leak */
+    col_unbind_iterator(iterator);
+
+
+    /* Bind iterator again in flat mode */
+    error =  col_bind_iterator(&iterator, peer, COL_TRAVERSE_DEFAULT | COL_TRAVERSE_END);
+    if (error) {
+        printf("Error (bind): %d\n", error);
+        col_destroy_collection(peer);
+        return error;
+    }
+
+    printf("\n\nCircled looping with pin (default):\n\n");
+
+    do {
+        /* Loop through a collection */
+        error = col_iterate_collection(iterator, &item);
+        if (error) {
+            printf("Error (iterate): %d\n", error);
+            col_destroy_collection(peer);
+            col_unbind_iterator(iterator);
+            return error;
+        }
+
+        if (strcmp(col_get_item_property(item, NULL), "queue") == 0) {
+            /* Make it a new looping point */
+            col_pin_iterator(iterator);
+            printf("Found pin point.\n\n");
+            break;
+        }
+        /* Are we done ? */
+        if (item == NULL) {
+            printf("Unexpected end.\n\n");
+            col_destroy_collection(peer);
+            col_unbind_iterator(iterator);
+            return EINVAL;
+        }
+        else {
+            depth = 0;
+            col_get_item_depth(iterator, &depth);
+            printf("%*s", depth * 4, "");
+            col_debug_item(item);
+        }
+    }
+    while(1);
+
+    /* Second loop around the pin point */
+    for (i = 0; i < 200; i++) {
+        /* Loop through a collection */
+        error = col_iterate_collection(iterator, &item);
+        if (error) {
+            printf("Error (iterate): %d\n", error);
+            col_destroy_collection(peer);
+            col_unbind_iterator(iterator);
+            return error;
+        }
+
+        /* Are we done ? */
+        if (item == NULL) printf("Reached end.\n\n");
+        else {
+            depth = 0;
+            col_get_item_depth(iterator, &depth);
+            printf("%*s", depth * 4, "");
+            col_debug_item(item);
+        }
+    }
+
+    /* Do not forget to unbind iterator - otherwise there will be a leak */
+    col_unbind_iterator(iterator);
+    col_destroy_collection(peer);
+
     return EOK;
 }
 
@@ -1264,9 +1537,9 @@ int search_test(void)
         (error = col_add_collection_to_collection(level1, NULL, NULL, level2, COL_ADD_MODE_REFERENCE)) ||
         (error = col_create_collection(&level3, "level3", 0)) ||
         (error = col_add_collection_to_collection(level1, "level2", NULL, level3, COL_ADD_MODE_REFERENCE)) ||
-        (error = col_create_collection(&level4, "level4", 0)) ||
+        (error = col_create_collection(&level4, "leveL4", 0)) ||
         (error = col_add_collection_to_collection(level1, "level3", NULL, level4, COL_ADD_MODE_REFERENCE)) ||
-        (error = col_add_int_property(level1, "level4", "id", 1)) ||
+        (error = col_add_int_property(level1, "leveL4", "id", 1)) ||
         (error = col_add_long_property(level1, "level3", "packets", 100000000L)) ||
         (error = col_add_binary_property(level1, "level2", "stack", binary_dump, sizeof(binary_dump)))) {
         col_destroy_collection(level1);
@@ -1365,6 +1638,146 @@ int search_test(void)
     return EOK;
 }
 
+/* Sort test */
+int sort_test(void)
+{
+    struct collection_item *level1 = NULL;
+    struct collection_item *level2a = NULL;
+    struct collection_item *level2b = NULL;
+    struct collection_item *level3 = NULL;
+    int error = 0;
+
+    printf("\n\n==== SORT TEST ====\n\n");
+
+    if ((error = col_create_collection(&level1, "level1", 0)) ||
+        (error = col_create_collection(&level2a, "level2a", 0)) ||
+        (error = col_add_collection_to_collection(level1, NULL, NULL, level2a, COL_ADD_MODE_REFERENCE)) ||
+        (error = col_create_collection(&level2b, "level2b", 0)) ||
+        (error = col_add_collection_to_collection(level1, NULL, NULL, level2b, COL_ADD_MODE_REFERENCE)) ||
+        (error = col_create_collection(&level3, "level3", 0)) ||
+        (error = col_add_collection_to_collection(level1, "level2a", NULL, level3, COL_ADD_MODE_REFERENCE)) ||
+        (error = col_add_collection_to_collection(level1, "level2b", NULL, level3, COL_ADD_MODE_REFERENCE)) ||
+        (error = col_add_int_property(level1, NULL, "int3", 1)) ||
+        (error = col_add_int_property(level1, NULL, "int2", 2)) ||
+        (error = col_add_int_property(level1, NULL, "int1", 3)) ||
+        (error = col_add_bool_property(level1, NULL, "bool3", 1)) ||
+        (error = col_add_bool_property(level1, NULL, "bool2", 1)) ||
+        (error = col_add_bool_property(level1, NULL, "bool1", 0)) ||
+        (error = col_add_unsigned_property(level1, NULL, "unsigned1", 2)) ||
+        (error = col_add_unsigned_property(level1, NULL, "unsigned3", 1)) ||
+        (error = col_add_unsigned_property(level1, NULL, "unsigned2", 3)) ||
+        (error = col_add_long_property(level1, NULL, "long3", 1)) ||
+        (error = col_add_long_property(level1, NULL, "long2", 2)) ||
+        (error = col_add_long_property(level1, NULL, "long1", 3)) ||
+        (error = col_add_ulong_property(level1, NULL, "ulong1", 2)) ||
+        (error = col_add_ulong_property(level1, NULL, "ulong3", 1)) ||
+        (error = col_add_ulong_property(level1, NULL, "ulong2", 3)) ||
+        (error = col_add_double_property(level1, NULL, "double1", 2.2)) ||
+        (error = col_add_double_property(level1, NULL, "double3", 1.1)) ||
+        (error = col_add_double_property(level1, NULL, "double2", 3.3)) ||
+        (error = col_add_int_property(level3, NULL, "int3L3", 1)) ||
+        (error = col_add_int_property(level3, NULL, "int2L3", 2)) ||
+        (error = col_add_int_property(level3, NULL, "int1L3", 3)) ||
+        (error = col_add_unsigned_property(level1, "level2a!level3", "unsigned1L3", 2)) ||
+        (error = col_add_unsigned_property(level1, "level2a!level3", "unsigned3L3", 1)) ||
+        (error = col_add_unsigned_property(level1, "level2a!level3", "unsigned2L3", 3)) ||
+        (error = col_add_long_property(level1, "level2b!level3", "long3L3", 1)) ||
+        (error = col_add_long_property(level1, "level2b!level3", "long2L3", 2)) ||
+        (error = col_add_long_property(level1, "level2b!level3", "long1L3", 3)) ||
+        (error = col_add_ulong_property(level1, "level3", "ulong1L3", 2)) ||
+        (error = col_add_ulong_property(level1, "level3", "ulong3L3", 1)) ||
+        (error = col_add_ulong_property(level1, "level3", "ulong2L3", 3)) ||
+        (error = col_add_bool_property(level3, NULL, "bool3", 1)) ||
+        (error = col_add_bool_property(level3, NULL, "bool2", 1)) ||
+        (error = col_add_bool_property(level3, NULL, "bool1", 0)) ||
+        (error = col_add_double_property(level3, NULL, "double1L3", 2.2)) ||
+        (error = col_add_double_property(level3, NULL, "double3L3", 1.1)) ||
+        (error = col_add_double_property(level3, NULL, "double2L3", 3.3))) {
+        col_destroy_collection(level1);
+        col_destroy_collection(level2a);
+        col_destroy_collection(level2b);
+        col_destroy_collection(level3);
+        printf("Failed to build test. Error %d\n", error);
+        return error;
+    }
+
+    printf("\nUNSORTED COLLECTION\n\n");
+    col_debug_collection(level1, COL_TRAVERSE_DEFAULT);
+
+    error = col_sort_collection(level1, COL_CMPIN_PROP_EQU, COL_SORT_SUB | COL_SORT_MYSUB);
+    if (error) {
+        col_destroy_collection(level1);
+        col_destroy_collection(level2a);
+        col_destroy_collection(level2b);
+        col_destroy_collection(level3);
+        printf("Failed sort. Error %d\n", error);
+        return error;
+    }
+
+    printf("\nSORTED BUT SKIPPING REFERENCES\n\n");
+    col_debug_collection(level1, COL_TRAVERSE_DEFAULT);
+
+    error = col_sort_collection(level1, COL_CMPIN_PROP_EQU, COL_SORT_SUB);
+    if (error) {
+        col_destroy_collection(level1);
+        col_destroy_collection(level2a);
+        col_destroy_collection(level2b);
+        col_destroy_collection(level3);
+        printf("Failed sort. Error %d\n", error);
+        return error;
+    }
+
+    printf("\nSORTED BUT NOT SKIPPING REFERENCES\n\n");
+    col_debug_collection(level1, COL_TRAVERSE_DEFAULT);
+
+    error = col_sort_collection(level1, COL_CMPIN_DATA_LEN, COL_SORT_SUB | COL_SORT_DESC);
+    if (error) {
+        col_destroy_collection(level1);
+        col_destroy_collection(level2a);
+        col_destroy_collection(level2b);
+        col_destroy_collection(level3);
+        printf("Failed sort. Error %d\n", error);
+        return error;
+    }
+
+    printf("\nSORTED DESC NOT SKIPPING BY LENGTH OF DATA\n\n");
+    col_debug_collection(level1, COL_TRAVERSE_DEFAULT);
+
+    error = col_sort_collection(level1, COL_CMPIN_PROP_LEN, COL_SORT_SUB | COL_SORT_DESC);
+    if (error) {
+        col_destroy_collection(level1);
+        col_destroy_collection(level2a);
+        col_destroy_collection(level2b);
+        col_destroy_collection(level3);
+        printf("Failed sort. Error %d\n", error);
+        return error;
+    }
+
+    printf("\nSORTED DESC NOT SKIPPING BY LENGTH OF PROPERTY\n\n");
+    col_debug_collection(level1, COL_TRAVERSE_DEFAULT);
+
+    error = col_sort_collection(level1, COL_CMPIN_DATA, COL_SORT_SUB | COL_SORT_DESC);
+    if (error) {
+        col_destroy_collection(level1);
+        col_destroy_collection(level2a);
+        col_destroy_collection(level2b);
+        col_destroy_collection(level3);
+        printf("Failed sort. Error %d\n", error);
+        return error;
+    }
+
+    printf("\nSORTED DESC NOT SKIPPING BY DATA\n\n");
+    col_debug_collection(level1, COL_TRAVERSE_DEFAULT);
+
+    col_destroy_collection(level1);
+    col_destroy_collection(level2a);
+    col_destroy_collection(level2b);
+    col_destroy_collection(level3);
+
+    printf("\n\n==== SORT TEST END ====\n\n");
+
+    return EOK;
+}
 
 /* Main function of the unit test */
 
@@ -1380,7 +1793,8 @@ int main(int argc, char *argv[])
         (error = iterator_test()) ||
         (error = insert_extract_test()) ||
         (error = delete_test()) ||
-        (error = search_test())) {
+        (error = search_test()) ||
+        (error = sort_test())) {
         printf("Failed!\n");
     }
     else printf("Success!\n");
