@@ -27,12 +27,12 @@
 #include <ctype.h>
 #include <stdlib.h>
 #include <locale.h>
+#include <fcntl.h>
 #include "config.h"
 /* For error text */
 #include <libintl.h>
 #define _(String) gettext (String)
 /* INI file is used as a collection */
-#include "collection_priv.h"
 #include "collection.h"
 #include "collection_tools.h"
 #include "trace.h"
@@ -76,6 +76,22 @@
 #define INI_ERROR       "errors"
 #define INI_ERROR_NAME  "errname"
 
+
+/* Internal sizes. MAX_KEY is defined in config.h */
+#define MAX_VALUE       PATH_MAX
+#define BUFFER_SIZE     MAX_KEY + MAX_VALUE + 3
+
+/* Internally used functions */
+static int config_with_lines(const char *application,
+                             FILE *config_file,
+                             const char *config_source,
+                             struct collection_item **ini_config,
+                             int error_level,
+                             struct collection_item **error_list,
+                             struct collection_item **lines);
+
+
+
 /* Different error string functions can be passed as callbacks */
 typedef const char * (*error_fn)(int error);
 
@@ -99,11 +115,15 @@ const char *parsing_error_str(int parsing_error)
             return str_error[parsing_error-1];
 }
 
-/* Function to return grammar error */
+/* Function to return grammar error.
+ * This function is currently not used.
+ * It is planned to be used by the INI
+ * file grammar parser.
+ */
 const char *grammar_error_str(int grammar_error)
 {
     const char *placeholder= _("Unknown grammar error.");
-    /* FIXME - this is a temporary placeholder !!!! */
+    /* THIS IS A TEMPORARY PLACEHOLDER !!!! */
     const char *str_error[] = { _(""),
                                 _(""),
                                 _(""),
@@ -120,11 +140,15 @@ const char *grammar_error_str(int grammar_error)
             return str_error[grammar_error-1];
 }
 
-/* Function to return validation error */
+/* Function to return validation error.
+ * This function is currently not used.
+ * It is planned to be used by the INI
+ * file grammar validator.
+ */
 const char *validation_error_str(int validation_error)
 {
     const char *placeholder= _("Unknown validation error.");
-    /* FIXME - this is a temporary placeholder !!!! */
+    /* THIS IS A TEMPORARY PLACEHOLDER !!!! */
     const char *str_error[] = { _(""),
                                 _(""),
                                 _(""),
@@ -154,13 +178,13 @@ int read_line(FILE *file,
 /***************************************************************************/
 /* Function to read single ini file and pupulate
  * the provided collection with subcollcetions from the file */
-static int ini_to_collection(const char *filename,
+static int ini_to_collection(FILE *file,
+                             const char *config_filename,
                              struct collection_item *ini_config,
                              int error_level,
                              struct collection_item **error_list,
                              struct collection_item **lines)
 {
-    FILE *file;
     int error;
     int status;
     int section_count = 0;
@@ -177,12 +201,9 @@ static int ini_to_collection(const char *filename,
 
     TRACE_FLOW_STRING("ini_to_collection", "Entry");
 
-    /* Open file for reading */
-    file = fopen(filename, "r");
     if (file == NULL) {
-        error = errno;
-        TRACE_ERROR_NUMBER("Failed to open file - but this is OK", error);
-        return ENOENT;
+        TRACE_ERROR_NUMBER("No file handle", EINVAL);
+        return EINVAL;
     }
 
     /* Open the collection of errors */
@@ -195,7 +216,7 @@ static int ini_to_collection(const char *filename,
             return error;
         }
         /* Add file name as the first item */
-        error = col_add_str_property(*error_list, NULL, INI_ERROR_NAME, filename, 0);
+        error = col_add_str_property(*error_list, NULL, INI_ERROR_NAME, config_filename, 0);
         if (error) {
             TRACE_ERROR_NUMBER("Failed to and name to collection", error);
             fclose(file);
@@ -424,9 +445,10 @@ void free_ini_config_lines(struct collection_item *lines)
     TRACE_FLOW_STRING("free_ini_config_lines", "Exit");
 }
 
+
 /* Read configuration information from a file */
 int config_from_file(const char *application,
-                     const char *config_file,
+                     const char *config_filename,
                      struct collection_item **ini_config,
                      int error_level,
                      struct collection_item **error_list)
@@ -435,7 +457,7 @@ int config_from_file(const char *application,
 
     TRACE_FLOW_STRING("config_from_file", "Entry");
     error = config_from_file_with_lines(application,
-                                        config_file,
+                                        config_filename,
                                         ini_config,
                                         error_level,
                                         error_list,
@@ -444,15 +466,96 @@ int config_from_file(const char *application,
     return error;
 }
 
+/* Read configuration information from a file descriptor */
+int config_from_fd(const char *application,
+                   int fd,
+                   const char *config_source,
+                   struct collection_item **ini_config,
+                   int error_level,
+                   struct collection_item **error_list)
+{
+    int error;
+
+    TRACE_FLOW_STRING("config_from_fd", "Entry");
+    error = config_from_fd_with_lines(application, fd, config_source,
+                                      ini_config, error_level,
+                                      error_list, NULL);
+    TRACE_FLOW_NUMBER("config_from_fd. Returns", error);
+    return error;
+}
+
 /* Function to read the ini file and have a collection
  * of which item appers on which line
  */
 int config_from_file_with_lines(const char *application,
-                                const char *config_file,
+                                const char *config_filename,
                                 struct collection_item **ini_config,
                                 int error_level,
                                 struct collection_item **error_list,
                                 struct collection_item **lines)
+{
+    int error = EOK;
+    FILE *config_file = NULL;
+
+    TRACE_FLOW_STRING("config_from_file_with_lines", "Entry");
+
+    config_file = fopen(config_filename, "r");
+    if(!config_file) {
+        error = errno;
+        TRACE_ERROR_NUMBER("Failed to open file", error);
+        return error;
+    }
+
+    error = config_with_lines(application, config_file,
+                              config_filename, ini_config,
+                              error_level, error_list,
+                              lines);
+    TRACE_FLOW_NUMBER("config_from_file_with_lines. Returns", error);
+    return error;
+}
+
+/* Function to read the ini file and have a collection
+ * of which item appers on which line
+ */
+int config_from_fd_with_lines(const char *application,
+                              int fd,
+                              const char *config_source,
+                              struct collection_item **ini_config,
+                              int error_level,
+                              struct collection_item **error_list,
+                              struct collection_item **lines)
+{
+    int error = EOK;
+    FILE *config_file;
+
+    TRACE_FLOW_STRING("config_from_fd_with_lines", "Entry");
+
+    config_file = fdopen(fd, "r");
+    if (!config_file) {
+        error = errno;
+        TRACE_ERROR_NUMBER("Failed to dup file", error);
+        return error;
+    }
+
+    error = config_with_lines(application, config_file,
+                              config_source, ini_config,
+                              error_level, error_list,
+                              lines);
+    TRACE_FLOW_NUMBER("config_from_fd_with_lines. Returns", error);
+
+    return error;
+}
+
+/* Low level function that prepares the collection
+ * and calls parser.
+ */
+static int config_with_lines(const char *application,
+                             FILE *config_file,
+                             const char *config_source,
+                             struct collection_item **ini_config,
+                             int error_level,
+                             struct collection_item **error_list,
+                             struct collection_item **lines)
 {
     int error;
     int created = 0;
@@ -512,8 +615,9 @@ int config_from_file_with_lines(const char *application,
     }
 
     /* Do the actual work */
-    error = ini_to_collection(config_file, *ini_config,
-                              error_level, error_list, lines);
+    error = ini_to_collection(config_file, config_source,
+                              *ini_config, error_level,
+                              error_list, lines);
     /* In case of error when we created collection - delete it */
     if (error && created) {
         col_destroy_collection(*ini_config);
@@ -529,6 +633,39 @@ int config_from_file_with_lines(const char *application,
     return error;
 }
 
+/* Special wrapper around the inernal parser
+ * to open the file first.
+ * Used in conf_for_app function.
+ */
+static int ini_to_col_from_file(const char *config_filename,
+                                struct collection_item *ini_config,
+                                int error_level,
+                                struct collection_item **error_list,
+                                struct collection_item **lines)
+{
+    int error = EOK;
+    FILE *config_file = NULL;
+
+    TRACE_FLOW_STRING("ini_to_col_from_file", "Entry");
+
+    config_file = fopen(config_filename, "r");
+    if(!config_file) {
+        error = errno;
+        TRACE_ERROR_NUMBER("ini_to_col_from_file. Returns", error);
+        return ENOENT;
+    }
+
+    error = ini_to_collection(config_file,
+                              config_filename,
+                              ini_config,
+                              error_level,
+                              error_list,
+                              lines);
+    TRACE_FLOW_NUMBER("ini_to_col_from_file. Returns", error);
+    return error;
+}
+
+
 /* Read default config file and then overwrite it with a specific one
  * from the directory */
 int config_for_app(const char *application,
@@ -545,11 +682,18 @@ int config_for_app(const char *application,
     struct collection_item **pass_common = NULL;
     struct collection_item **pass_specific = NULL;
     int created = 0;
+    int tried = 0;
+    int noents = 0;
 
     TRACE_FLOW_STRING("config_to_collection", "Entry");
 
     if (ini_config == NULL) {
-        TRACE_ERROR_NUMBER("Failed to create collection", EINVAL);
+        TRACE_ERROR_NUMBER("Invalid parameter", EINVAL);
+        return EINVAL;
+    }
+
+    if ((config_file == NULL) && (config_dir == NULL)) {
+        TRACE_ERROR_NUMBER("Noop call of the function is invalid", EINVAL);
         return EINVAL;
     }
 
@@ -582,10 +726,13 @@ int config_for_app(const char *application,
                                       COL_CLASS_INI_CONFIG);
         if (error != EOK) {
             TRACE_ERROR_NUMBER("Failed to create collection", error);
-            col_destroy_collection(*error_set);
-            *error_set = NULL;
+            if (error_set) {
+                col_destroy_collection(*error_set);
+                *error_set = NULL;
+            }
             return error;
         }
+        created = 1;
     }
     /* Is the collection of the right class? */
     else if (col_is_of_class(*ini_config, COL_CLASS_INI_CONFIG)) {
@@ -596,22 +743,26 @@ int config_for_app(const char *application,
     /* Read master file */
     if (config_file != NULL) {
         TRACE_INFO_STRING("Reading master file:", config_file);
-        error = ini_to_collection(config_file, *ini_config,
-                                  error_level, pass_common, NULL);
+        error = ini_to_col_from_file(config_file, *ini_config,
+                                     error_level, pass_common, NULL);
+        tried++;
         /* ENOENT and EOK are Ok */
-        if (error && (error != ENOENT)) {
-            TRACE_ERROR_NUMBER("Failed to read master file", error);
-            /* In case of error when we created collection - delete it */
-            if(error && created) {
-                col_destroy_collection(*ini_config);
-                *ini_config = NULL;
+        if (error) {
+            if (error != ENOENT) {
+                TRACE_ERROR_NUMBER("Failed to read master file", error);
+                /* In case of error when we created collection - delete it */
+                if(error && created) {
+                    col_destroy_collection(*ini_config);
+                    *ini_config = NULL;
+                }
+                /* We do not clear the error_set here */
+                return error;
             }
-            /* We do not clear the error_set here */
-            return error;
+            else noents++;
         }
         /* Add error results if any to the overarching error collection */
         if ((pass_common != NULL) && (*pass_common != NULL)) {
-            TRACE_INFO_STRING("Process erros resulting from file:", config_file);
+            TRACE_INFO_STRING("Process errors resulting from file:", config_file);
             error = col_add_collection_to_collection(*error_set, NULL, NULL,
                                                      *pass_common,
                                                      COL_ADD_MODE_EMBED);
@@ -620,8 +771,10 @@ int config_for_app(const char *application,
                     col_destroy_collection(*ini_config);
                     *ini_config = NULL;
                 }
-                col_destroy_collection(*error_set);
-                *error_set = NULL;
+                if (error_set) {
+                    col_destroy_collection(*error_set);
+                    *error_set = NULL;
+                }
                 TRACE_ERROR_NUMBER("Failed to add error collection to another error collection", error);
                 return error;
             }
@@ -635,34 +788,38 @@ int config_for_app(const char *application,
             error = ENOMEM;
             TRACE_ERROR_NUMBER("Failed to allocate memory for file name", error);
             /* In case of error when we created collection - delete it */
-            if(error && created) {
+            if(created) {
                 col_destroy_collection(*ini_config);
                 *ini_config = NULL;
             }
-            col_destroy_collection(*error_set);
-            *error_set = NULL;
+            if (error_set) {
+                col_destroy_collection(*error_set);
+                *error_set = NULL;
+            }
             return error;
         }
 
         sprintf(file_name, "%s%s%s.conf", config_dir, SLASH, application);
         TRACE_INFO_STRING("Opening file:", file_name);
-
-        /* Read master file */
-        error = ini_to_collection(file_name, *ini_config,
-                                  error_level, pass_specific, NULL);
+        /* Read specific file */
+        error = ini_to_col_from_file(file_name, *ini_config,
+                                     error_level, pass_specific, NULL);
+        tried++;
         free(file_name);
         /* ENOENT and EOK are Ok */
-        if (error && (error != ENOENT)) {
-            TRACE_ERROR_NUMBER("Failed to read specific application file", error);
-            /* In case of error when we created collection - delete it */
-            if (error && created) {
-                col_destroy_collection(*ini_config);
-                *ini_config = NULL;
+        if (error) {
+            if (error != ENOENT) {
+                TRACE_ERROR_NUMBER("Failed to read specific application file", error);
+                /* In case of error when we created collection - delete it */
+                if (error && created) {
+                    col_destroy_collection(*ini_config);
+                    *ini_config = NULL;
+                }
+                /* We do not clear the error_set here */
+                return error;
             }
-            /* We do not clear the error_set here */
-            return error;
+            else noents++;
         }
-
         /* Add error results if any to the overarching error collection */
         if ((pass_specific != NULL) && (*pass_specific != NULL)) {
             error = col_add_collection_to_collection(*error_set, NULL, NULL,
@@ -673,14 +830,36 @@ int config_for_app(const char *application,
                     col_destroy_collection(*ini_config);
                     *ini_config = NULL;
                 }
-                col_destroy_collection(*error_set);
-                *error_set = NULL;
+                if (error_set) {
+                    col_destroy_collection(*error_set);
+                    *error_set = NULL;
+                }
                 TRACE_ERROR_NUMBER("Failed to add error collection to another error collection", error);
                 return error;
             }
         }
     }
 
+    /* If we failed to read or access file as many
+     * times as we tried and we told to stop on any errors
+     * we should report an error.
+     */
+    TRACE_INFO_NUMBER("Tried:", tried);
+    TRACE_INFO_NUMBER("Noents:", noents);
+
+    if ((tried == noents) && (error_level == INI_STOP_ON_ANY)) {
+        TRACE_ERROR_NUMBER("Fail to read or access all the files tried", ENOENT);
+        if (created) {
+            col_destroy_collection(*ini_config);
+            *ini_config = NULL;
+        }
+        if (error_set) {
+            col_destroy_collection(*error_set);
+            *error_set = NULL;
+        }
+        return ENOENT;
+    }
+
     TRACE_FLOW_STRING("config_to_collection", "Exit");
     return EOK;
 }
@@ -1432,9 +1611,9 @@ char **get_string_config_array(struct collection_item *item,
 
     /* Loop through the string */
     dest = copy;
-    buff = item->data;
+    buff = col_get_item_data(item);
     start = buff;
-    dlen = item->length - 1;
+    dlen = col_get_item_length(item) - 1;
     for(i = 0; i < dlen; i++) {
         growlen = 1;
         for(j = 0; j < lensep; j++) {
