@@ -21,19 +21,14 @@
 
 #define _GNU_SOURCE
 
+#include <ctype.h>
 #include "config.h"
 #include "util/util.h"
 #include "confdb/confdb.h"
 #include "confdb/confdb_private.h"
-#include "util/btreemap.h"
+#include "util/strtonum.h"
 #include "db/sysdb.h"
 
-#define CONFDB_DOMAINS_PATH "config/domains"
-#define CONFDB_DOMAIN_BASEDN "cn=domains,cn=config"
-#define CONFDB_DOMAIN_ATTR "cn"
-#define CONFDB_MPG "magicPrivateGroups"
-#define CONFDB_FQ "useFullyQualifiedNames"
-
 #define CONFDB_ZERO_CHECK_OR_JUMP(var, ret, err, label) do { \
     if (!var) { \
         ret = err; \
@@ -108,110 +103,6 @@ done:
     return ret;
 }
 
-/* split a string into an allocated array of strings.
- * the separator is a string, and is case-sensitive.
- * optionally single values can be trimmed of of spaces and tabs */
-static int split_on_separator(TALLOC_CTX *mem_ctx, const char *str,
-                              const char *sep, bool trim, char ***_list, int *size)
-{
-    const char *t, *p, *n;
-    size_t l, s, len;
-    char **list, **r;
-
-    if (!str || !*str || !sep || !*sep || !_list) return EINVAL;
-
-    s = strlen(sep);
-    t = str;
-
-    list = NULL;
-    l = 0;
-
-    if (trim)
-        while (*t == ' ' || *t == '\t') t++;
-
-    while (t && (p = strstr(t, sep))) {
-        len = p - t;
-        n = p + s; /* save next string starting point */
-        if (trim) {
-            while (*t == ' ' || *t == '\t') {
-                t++;
-                len--;
-                if (len == 0) break;
-            }
-            p--;
-            while (len > 0 && (*p == ' ' || *p == '\t')) {
-                len--;
-                p--;
-            }
-        }
-
-        r = talloc_realloc(mem_ctx, list, char *, l + 2);
-        if (!r) {
-            talloc_free(list);
-            return ENOMEM;
-        } else {
-            list = r;
-        }
-
-        if (len == 0) {
-            list[l] = talloc_strdup(list, "");
-        } else {
-            list[l] = talloc_strndup(list, t, len);
-        }
-        if (!list[l]) {
-            talloc_free(list);
-            return ENOMEM;
-        }
-        l++;
-
-        t = n; /* move to next string */
-    }
-
-    if (t) {
-        r = talloc_realloc(mem_ctx, list, char *, l + 2);
-        if (!r) {
-            talloc_free(list);
-            return ENOMEM;
-        } else {
-            list = r;
-        }
-
-        if (trim) {
-            len = strlen(t);
-            while (*t == ' ' || *t == '\t') {
-                t++;
-                len--;
-                if (len == 0) break;
-            }
-            p = t + len - 1;
-            while (len > 0 && (*p == ' ' || *p == '\t')) {
-                len--;
-                p--;
-            }
-
-            if (len == 0) {
-                list[l] = talloc_strdup(list, "");
-            } else {
-                list[l] = talloc_strndup(list, t, len);
-            }
-        } else {
-            list[l] = talloc_strdup(list, t);
-        }
-        if (!list[l]) {
-            talloc_free(list);
-            return ENOMEM;
-        }
-        l++;
-    }
-
-    list[l] = NULL; /* terminate list */
-
-    if (size) *size = l + 1;
-    *_list = list;
-
-    return EOK;
-}
-
 int confdb_add_param(struct confdb_ctx *cdb,
                      bool replace,
                      const char *section,
@@ -613,7 +504,7 @@ int confdb_get_string_as_list(struct confdb_ctx *cdb, TALLOC_CTX *ctx,
         goto done;
     }
 
-    ret = split_on_separator(ctx, values[0], ",", true, result, NULL);
+    ret = split_on_separator(ctx, values[0], ',', true, result, NULL);
 
 done:
     talloc_free(values);
@@ -625,7 +516,6 @@ done:
 }
 
 int confdb_init(TALLOC_CTX *mem_ctx,
-                struct tevent_context *ev,
                 struct confdb_ctx **cdb_ctx,
                 char *confdb_location)
 {
@@ -664,6 +554,8 @@ int confdb_init(TALLOC_CTX *mem_ctx,
 
     ret = ldb_connect(cdb->ldb, confdb_location, 0, NULL);
     if (ret != LDB_SUCCESS) {
+        DEBUG(0, ("Unable to open config database [%s]\n",
+                  confdb_location));
         talloc_free(cdb);
         return EIO;
     }
@@ -673,17 +565,88 @@ int confdb_init(TALLOC_CTX *mem_ctx,
     return EOK;
 }
 
-int confdb_get_domain(struct confdb_ctx *cdb,
-                      TALLOC_CTX *mem_ctx,
-                      const char *name,
-                      struct sss_domain_info **_domain)
+static errno_t get_entry_as_uint32(struct ldb_message *msg,
+                                   uint32_t *return_value,
+                                   const char *entry,
+                                   uint32_t default_value)
+{
+    const char *tmp = NULL;
+    char *endptr;
+    uint32_t u32ret = 0;
+
+    *return_value = 0;
+
+    if (!msg || !entry) {
+        return EFAULT;
+    }
+
+    tmp = ldb_msg_find_attr_as_string(msg, entry, NULL);
+    if (tmp == NULL) {
+        *return_value = default_value;
+        return EOK;
+    }
+
+    if ((*tmp == '-') || (*tmp == '\0')) {
+        return EINVAL;
+    }
+
+    u32ret = strtouint32 (tmp, &endptr, 10);
+    if (errno) {
+        return errno;
+    }
+
+    if (*endptr != '\0') {
+        /* Not all of the string was a valid number */
+        return EINVAL;
+    }
+
+    *return_value = u32ret;
+    return EOK;
+}
+
+static errno_t get_entry_as_bool(struct ldb_message *msg,
+                                   bool *return_value,
+                                   const char *entry,
+                                   bool default_value)
+{
+    const char *tmp = NULL;
+
+    *return_value = 0;
+
+    if (!msg || !entry) {
+        return EFAULT;
+    }
+
+    tmp = ldb_msg_find_attr_as_string(msg, entry, NULL);
+    if (tmp == NULL || *tmp == '\0') {
+        *return_value = default_value;
+        return EOK;
+    }
+
+    if (strcasecmp(tmp, "FALSE") == 0) {
+        *return_value = 0;
+    }
+    else if (strcasecmp(tmp, "TRUE") == 0) {
+        *return_value = 1;
+    }
+    else {
+        return EINVAL;
+    }
+
+    return EOK;
+}
+
+static int confdb_get_domain_internal(struct confdb_ctx *cdb,
+                                      TALLOC_CTX *mem_ctx,
+                                      const char *name,
+                                      struct sss_domain_info **_domain)
 {
     struct sss_domain_info *domain;
     struct ldb_result *res;
     TALLOC_CTX *tmp_ctx;
     struct ldb_dn *dn;
     const char *tmp;
-    int ret;
+    int ret, val;
 
     tmp_ctx = talloc_new(mem_ctx);
     if (!tmp_ctx) return ENOMEM;
@@ -726,7 +689,9 @@ int confdb_get_domain(struct confdb_ctx *cdb,
         goto done;
     }
 
-    tmp = ldb_msg_find_attr_as_string(res->msgs[0], "provider", NULL);
+    tmp = ldb_msg_find_attr_as_string(res->msgs[0],
+                                      CONFDB_DOMAIN_ID_PROVIDER,
+                                      NULL);
     if (tmp) {
         domain->provider = talloc_strdup(domain, tmp);
         if (!domain->provider) {
@@ -735,53 +700,124 @@ int confdb_get_domain(struct confdb_ctx *cdb,
         }
     }
     else {
-        DEBUG(0, ("Domain [%s] does not specify a provider, disabling!\n",
+        DEBUG(0, ("Domain [%s] does not specify an ID provider, disabling!\n",
                   domain->name));
         ret = EINVAL;
+        goto done;
+    }
+
+    if (strcasecmp(domain->provider, "files") == 0) {
+        /* The files provider is not valid anymore */
+        DEBUG(0, ("The \"files\" provider is invalid\n"));
+        ret = EINVAL;
+        goto done;
+    }
+
+    if (strcasecmp(domain->provider, "local") == 0) {
+        /* If this is the local provider, we need to ensure that
+         * no other provider was specified for other types, since
+         * the local provider cannot load them.
+         */
+        tmp = ldb_msg_find_attr_as_string(res->msgs[0],
+                                          CONFDB_DOMAIN_AUTH_PROVIDER,
+                                          NULL);
+        if (tmp && strcasecmp(tmp, "local") != 0) {
+            DEBUG(0, ("Local ID provider does not support [%s] as an AUTH provider.\n", tmp));
+            ret = EINVAL;
+            goto done;
+        }
+
+        tmp = ldb_msg_find_attr_as_string(res->msgs[0],
+                                          CONFDB_DOMAIN_ACCESS_PROVIDER,
+                                          NULL);
+        if (tmp && strcasecmp(tmp, "local") != 0) {
+            DEBUG(0, ("Local ID provider does not support [%s] as an ACCESS provider.\n", tmp));
+            ret = EINVAL;
+            goto done;
+        }
+
+        tmp = ldb_msg_find_attr_as_string(res->msgs[0],
+                                          CONFDB_DOMAIN_CHPASS_PROVIDER,
+                                          NULL);
+        if (tmp && strcasecmp(tmp, "local") != 0) {
+            DEBUG(0, ("Local ID provider does not support [%s] as a CHPASS provider.\n", tmp));
+            ret = EINVAL;
+            goto done;
+        }
     }
 
     domain->timeout = ldb_msg_find_attr_as_int(res->msgs[0],
-                                               "timeout", 0);
+                                               CONFDB_DOMAIN_TIMEOUT, 0);
 
     /* Determine if this domain can be enumerated */
-    domain->enumerate = ldb_msg_find_attr_as_int(res->msgs[0],
-                                                 "enumerate", 0);
-    if (domain->enumerate == 0) {
+
+    /* TEMP: test if the old bitfield conf value is used and warn it has been
+     * superceeded. */
+    val = ldb_msg_find_attr_as_int(res->msgs[0], CONFDB_DOMAIN_ENUMERATE, 0);
+    if (val > 0) { /* ok there was a number in here */
+        DEBUG(0, ("Warning: enumeration parameter in %s still uses integers! "
+                  "Enumeration is now a boolean and takes true/false values. "
+                  "Interpreting as true\n", domain->name));
+        domain->enumerate = true;
+    } else { /* assume the new format */
+        ret = get_entry_as_bool(res->msgs[0], &domain->enumerate,
+                                CONFDB_DOMAIN_ENUMERATE, 1);
+        if(ret != EOK) {
+            DEBUG(0, ("Invalid value for %s\n", CONFDB_DOMAIN_ENUMERATE));
+            goto done;
+        }
+    }
+    if (!domain->enumerate) {
         DEBUG(1, ("No enumeration for [%s]!\n", domain->name));
     }
 
-    /* Determine if this is a legacy domain */
-    if (ldb_msg_find_attr_as_bool(res->msgs[0], "legacy", 0)) {
-        domain->legacy = true;
-    }
-
-    /* Determine if this is domain uses MPG */
-    if (strcasecmp(domain->provider, "local") == 0 ||
-        ldb_msg_find_attr_as_bool(res->msgs[0], CONFDB_MPG, 0)) {
-        domain->mpg = true;
-    }
-
     /* Determine if user/group names will be Fully Qualified
      * in NSS interfaces */
-    if (ldb_msg_find_attr_as_bool(res->msgs[0], CONFDB_FQ, 0)) {
-        domain->fqnames = true;
+    ret = get_entry_as_bool(res->msgs[0], &domain->fqnames, CONFDB_DOMAIN_FQ, 0);
+    if(ret != EOK) {
+        DEBUG(0, ("Invalid value for %s\n", CONFDB_DOMAIN_FQ));
+        goto done;
     }
 
-    domain->id_min = ldb_msg_find_attr_as_uint(res->msgs[0],
-                                               "minId", SSSD_MIN_ID);
-    domain->id_max = ldb_msg_find_attr_as_uint(res->msgs[0],
-                                               "maxId", 0);
+    ret = get_entry_as_uint32(res->msgs[0], &domain->id_min,
+                              CONFDB_DOMAIN_MINID, SSSD_MIN_ID);
+    if (ret != EOK) {
+        DEBUG(0, ("Invalid value for minId\n"));
+        ret = EINVAL;
+        goto done;
+    }
+
+    ret = get_entry_as_uint32(res->msgs[0], &domain->id_max,
+                              CONFDB_DOMAIN_MAXID, 0);
+    if (ret != EOK) {
+        DEBUG(0, ("Invalid value for maxId\n"));
+        ret = EINVAL;
+        goto done;
+    }
+
+    if (domain->id_max && (domain->id_max < domain->id_min)) {
+        DEBUG(0, ("Invalid domain range\n"));
+        ret = EINVAL;
+        goto done;
+    }
 
     /* Do we allow to cache credentials */
-    if (ldb_msg_find_attr_as_bool(res->msgs[0], "cache-credentials", 0)) {
-        domain->cache_credentials = true;
+    ret = get_entry_as_bool(res->msgs[0], &domain->cache_credentials,
+                            CONFDB_DOMAIN_CACHE_CREDS, 0);
+    if(ret != EOK) {
+        DEBUG(0, ("Invalid value for %s\n", CONFDB_DOMAIN_CACHE_CREDS));
+        goto done;
     }
 
-    if (ldb_msg_find_attr_as_bool(res->msgs[0], "store-legacy-passwords", 0)) {
-        domain->legacy_passwords = true;
+    ret = get_entry_as_bool(res->msgs[0], &domain->legacy_passwords,
+                            CONFDB_DOMAIN_LEGACY_PASS, 0);
+    if(ret != EOK) {
+        DEBUG(0, ("Invalid value for %s\n", CONFDB_DOMAIN_LEGACY_PASS));
+        goto done;
     }
 
     *_domain = domain;
+    ret = EOK;
 
 done:
     talloc_free(tmp_ctx);
@@ -789,20 +825,25 @@ done:
 }
 
 int confdb_get_domains(struct confdb_ctx *cdb,
-                       TALLOC_CTX *mem_ctx,
                        struct sss_domain_info **domains)
 {
     TALLOC_CTX *tmp_ctx;
     struct sss_domain_info *domain, *prevdom = NULL;
-    struct sss_domain_info *first = NULL;
     char **domlist;
     int ret, i;
 
-    tmp_ctx = talloc_new(mem_ctx);
+    if (cdb->doms) {
+        *domains = cdb->doms;
+        return EOK;
+    }
+
+    tmp_ctx = talloc_new(NULL);
     if (!tmp_ctx) return ENOMEM;
 
     ret = confdb_get_string_as_list(cdb, tmp_ctx,
-                                    CONFDB_DOMAINS_PATH, "domains", &domlist);
+                                    CONFDB_MONITOR_CONF_ENTRY,
+                                    CONFDB_MONITOR_ACTIVE_DOMAINS,
+                                    &domlist);
     if (ret == ENOENT) {
         DEBUG(0, ("No domains configured, fatal error!\n"));
         goto done;
@@ -813,7 +854,7 @@ int confdb_get_domains(struct confdb_ctx *cdb,
     }
 
     for (i = 0; domlist[i]; i++) {
-        ret = confdb_get_domain(cdb, mem_ctx, domlist[i], &domain);
+        ret = confdb_get_domain_internal(cdb, cdb, domlist[i], &domain);
         if (ret) {
             DEBUG(0, ("Error (%d [%s]) retrieving domain [%s], skipping!\n",
                       ret, strerror(ret), domlist[i]));
@@ -821,23 +862,47 @@ int confdb_get_domains(struct confdb_ctx *cdb,
             continue;
         }
 
-        if (first == NULL) {
-            first = domain;
-            prevdom = first;
+        if (cdb->doms == NULL) {
+            cdb->doms = domain;
+            prevdom = cdb->doms;
         } else {
             prevdom->next = domain;
             prevdom = domain;
         }
     }
 
-    if (first == NULL) {
-        DEBUG(0, ("No domains configured, fatal error!\n"));
+    if (cdb->doms == NULL) {
+        DEBUG(0, ("No properly configured domains, fatal error!\n"));
         ret = ENOENT;
+        goto done;
     }
 
-    *domains = first;
+    *domains = cdb->doms;
+    ret = EOK;
 
 done:
     talloc_free(tmp_ctx);
     return ret;
 }
+
+int confdb_get_domain(struct confdb_ctx *cdb,
+                      const char *name,
+                      struct sss_domain_info **_domain)
+{
+    struct sss_domain_info *dom, *doms;
+    int ret;
+
+    ret = confdb_get_domains(cdb, &doms);
+    if (ret != EOK) {
+        return ret;
+    }
+
+    for (dom = doms; dom; dom = dom->next) {
+        if (strcasecmp(dom->name, name) == 0) {
+            *_domain = dom;
+            return EOK;
+        }
+    }
+
+    return ENOENT;
+}
