@@ -56,14 +56,12 @@
 /* ping time cannot be less then once every few seconds or the
  * monitor will get crazy hammering children with messages */
 #define MONITOR_DEF_PING_TIME 10
-#define MONITOR_CONF_ENTRY "config/services/monitor"
 
 struct svc_spy;
 
 struct mt_svc {
     struct mt_svc *prev;
     struct mt_svc *next;
-
     struct sbus_connection *conn;
     struct svc_spy *conn_spy;
 
@@ -77,9 +75,12 @@ struct mt_svc {
 
     int ping_time;
 
+    bool svc_started;
+
     int restarts;
     time_t last_restart;
-    time_t last_pong;
+    time_t last_ping;
+    int failed_pongs;
 
     int debug_level;
 
@@ -116,9 +117,11 @@ struct mt_ctx {
     struct config_file_ctx *file_ctx;
     int inotify_fd;
     int service_id_timeout;
+    bool check_children;
+    bool services_started;
 };
 
-static int start_service(struct mt_svc *mt_svc, bool startup);
+static int start_service(struct mt_svc *mt_svc);
 
 static int monitor_service_init(struct sbus_connection *conn, void *data);
 
@@ -135,11 +138,16 @@ static int get_service_config(struct mt_ctx *ctx, const char *name,
                               struct mt_svc **svc_cfg);
 static int get_provider_config(struct mt_ctx *ctx, const char *name,
                               struct mt_svc **svc_cfg);
-static int add_new_service(struct mt_ctx *ctx, const char *name, bool startup);
-static int add_new_provider(struct mt_ctx *ctx, const char *name, bool startup);
+static int add_new_service(struct mt_ctx *ctx, const char *name);
+static int add_new_provider(struct mt_ctx *ctx, const char *name);
 
+static int mark_service_as_started(struct mt_svc *svc);
+
+#if 0
 static int monitor_signal_reconf(struct config_file_ctx *file_ctx,
                                  const char *filename);
+#endif
+
 static int update_monitor_config(struct mt_ctx *ctx);
 static int monitor_cleanup(void);
 
@@ -239,12 +247,9 @@ static int client_registration(DBusMessage *message,
     /* Fill in svc structure with connection data */
     svc->conn = mini->conn;
 
-    /* we need to attach a spy to the connection structure so that if some code
-     * frees it we can zero it out in the service structure. Otherwise we may
-     * try to access or even free, freed memory. */
-    ret = add_svc_conn_spy(svc);
+    ret = mark_service_as_started(svc);
     if (ret) {
-        DEBUG(0, ("Failed to attch spy\n"));
+        DEBUG(1, ("Failed to mark service [%s]!\n", svc_name));
         goto done;
     }
 
@@ -283,6 +288,9 @@ static int svc_destructor(void *mem)
         return 0;
     }
 
+    /* always try to delist service */
+    DLIST_REMOVE(svc->mt_ctx->svc_list, svc);
+
     /* svc is beeing freed, neutralize the spy */
     if (svc->conn_spy) {
         talloc_set_destructor((TALLOC_CTX *)svc->conn_spy, NULL);
@@ -318,6 +326,93 @@ static int add_svc_conn_spy(struct mt_svc *svc)
     return EOK;
 }
 
+static int mark_service_as_started(struct mt_svc *svc)
+{
+    struct mt_ctx *ctx = svc->mt_ctx;
+    struct mt_svc *iter;
+    int ret;
+    int i;
+
+    DEBUG(5, ("Marking %s as started.\n", svc->name));
+    svc->svc_started = true;
+
+    /* we need to attach a spy to the connection structure so that if some code
+     * frees it we can zero it out in the service structure. Otherwise we may
+     * try to access or even free, freed memory. */
+    ret = add_svc_conn_spy(svc);
+    if (ret) {
+        DEBUG(0, ("Failed to attch spy\n"));
+        goto done;
+    }
+
+    if (!ctx->services_started) {
+
+        /* check if all providers are up */
+        for (iter = ctx->svc_list; iter; iter = iter->next) {
+            if (iter->provider && !iter->svc_started) {
+                DEBUG(5, ("Still waiting on %s provider.", iter->name));
+                break;
+            }
+        }
+
+        if (iter) {
+            /* there are still unstarted providers */
+            goto done;
+        }
+
+        ctx->services_started = true;
+
+        DEBUG(4, ("Now starting services!\n"));
+        /* then start all services */
+        for (i = 0; ctx->services[i]; i++) {
+            add_new_service(ctx, ctx->services[i]);
+        }
+    }
+
+done:
+    return ret;
+}
+
+static void services_startup_timeout(struct tevent_context *ev,
+                                     struct tevent_timer *te,
+                                     struct timeval t, void *ptr)
+{
+    struct mt_ctx *ctx = talloc_get_type(ptr, struct mt_ctx);
+    int i;
+
+    DEBUG(6, ("Handling timeout\n"));
+
+    if (!ctx->services_started) {
+
+        DEBUG(1, ("Providers did not start in time, "
+                  "forcing services startup!\n"));
+
+        ctx->services_started = true;
+
+        DEBUG(4, ("Now starting services!\n"));
+        /* then start all services */
+        for (i = 0; ctx->services[i]; i++) {
+            add_new_service(ctx, ctx->services[i]);
+        }
+    }
+}
+
+static int add_services_startup_timeout(struct mt_ctx *ctx)
+{
+    struct tevent_timer *to;
+    struct timeval tv;
+
+    /* 5 seconds should be plenty */
+    tv = tevent_timeval_current_ofs(5, 0);
+    to = tevent_add_timer(ctx->ev, ctx, tv, services_startup_timeout, ctx);
+    if (!to) {
+        DEBUG(0,("Out of memory?!\n"));
+        return ENOMEM;
+    }
+
+    return EOK;
+}
+
 struct sbus_method monitor_methods[] = {
     { MON_SRV_METHOD_VERSION, get_monitor_version },
     { MON_SRV_METHOD_REGISTER, client_registration },
@@ -339,10 +434,9 @@ static int monitor_dbus_init(struct mt_ctx *ctx)
     char *monitor_address;
     int ret;
 
-    monitor_address = talloc_asprintf(ctx, "unix:path=%s/%s",
-                                      PIPE_PATH, SSSD_SERVICE_PIPE);
-    if (!monitor_address) {
-        return ENOMEM;
+    ret = monitor_get_sbus_address(ctx, &monitor_address);
+    if (ret != EOK) {
+        return ret;
     }
 
     ret = sbus_new_server(ctx, ctx->ev,
@@ -378,7 +472,7 @@ static void svc_try_restart(struct mt_svc *svc, time_t now)
      */
     talloc_free(svc->ping_ev);
 
-    ret = start_service(svc, false);
+    ret = start_service(svc);
     if (ret != EOK) {
         DEBUG(0,("Failed to restart service '%s'\n", svc->name));
         talloc_free(svc);
@@ -434,8 +528,13 @@ static void tasks_check_handler(struct tevent_context *ev,
             break;
         }
 
-        if (svc->last_pong != 0) {
-            if ((now - svc->last_pong) > (svc->ping_time * 3)) {
+        if (svc->last_ping != 0) {
+            if ((now - svc->last_ping) > (svc->ping_time)) {
+                svc->failed_pongs++;
+            } else {
+                svc->failed_pongs = 0;
+            }
+            if (svc->failed_pongs > 3) {
                 /* too long since we last heard of this process */
                 DEBUG(1, ("Killing service [%s], not responding to pings!\n",
                           svc->name));
@@ -444,6 +543,7 @@ static void tasks_check_handler(struct tevent_context *ev,
             }
         }
 
+        svc->last_ping = now;
     }
 
     if (!process_alive) {
@@ -481,6 +581,10 @@ static void global_checks_handler(struct tevent_context *ev,
     int status;
     pid_t pid;
 
+    if (!ctx->check_children) {
+        goto done;
+    }
+
     errno = 0;
     pid = waitpid(0, &status, WNOHANG);
     if (pid == 0) {
@@ -653,6 +757,10 @@ done:
     dbus_message_unref(reply);
 }
 
+#if 0
+This function should be re-enabled once live configuration updates
+are working properly.
+
 static int monitor_signal_reconf(struct config_file_ctx *file_ctx,
                                  const char *filename)
 {
@@ -670,6 +778,7 @@ static int monitor_signal_reconf(struct config_file_ctx *file_ctx,
     /* Update the monitor's configuration and signal children */
     return update_monitor_config(file_ctx->mt_ctx);
 }
+#endif
 
 static int service_signal_dns_reload(struct mt_svc *svc);
 static int monitor_update_resolv(struct config_file_ctx *file_ctx,
@@ -779,33 +888,6 @@ static int check_domain_ranges(struct sss_domain_info *domains)
     return EOK;
 }
 
-static int append_data_provider(struct mt_ctx *ctx)
-{
-    int i;
-    char **new_services;
-
-    for (i = 0; ctx->services[i]; i++) {
-        if (strcasecmp(ctx->services[i], "dp") == 0) {
-            return EOK;
-        }
-    }
-
-    new_services = talloc_realloc(ctx, ctx->services, char *, i+2);
-    if (new_services == NULL) {
-        return ENOMEM;
-    }
-    ctx->services = new_services;
-
-    ctx->services[i] = talloc_asprintf(ctx, "dp");
-    if (ctx->services[i] == NULL) {
-        return ENOMEM;
-    }
-    ctx->services[i+1] = NULL;
-    DEBUG(4, ("Added mandatory service Data Provider\n"));
-
-    return EOK;
-}
-
 static int check_local_domain_unique(struct sss_domain_info *domains)
 {
     uint8_t count = 0;
@@ -831,13 +913,37 @@ static int check_local_domain_unique(struct sss_domain_info *domains)
     return EOK;
 }
 
+static char *check_services(char **services)
+{
+    const char *known_services[] = { "nss", "pam", NULL };
+    int i;
+    int ii;
+
+    /* Check if services we are about to start are in the list if known */
+    for (i = 0; services[i]; i++) {
+        for (ii=0; known_services[ii]; ii++) {
+            if (strcasecmp(services[i], known_services[ii]) == 0) {
+                break;
+            }
+        }
+
+        if (known_services[ii] == NULL) {
+            return services[i];
+        }
+    }
+
+    return NULL;
+}
+
 int get_monitor_config(struct mt_ctx *ctx)
 {
     int ret;
     int timeout_seconds;
+    char *badsrv = NULL;
 
     ret = confdb_get_int(ctx->cdb, ctx,
-                         MONITOR_CONF_ENTRY, "sbusTimeout",
+                         CONFDB_MONITOR_CONF_ENTRY,
+                         CONFDB_MONITOR_SBUS_TIMEOUT,
                          10, &timeout_seconds);
     if (ret != EOK) {
         return ret;
@@ -850,23 +956,25 @@ int get_monitor_config(struct mt_ctx *ctx)
         return ENOMEM;
     }
     ret = confdb_get_string_as_list(ctx->cdb, ctx->service_ctx,
-                                    SERVICE_CONF_ENTRY, "activeServices",
+                                    CONFDB_MONITOR_CONF_ENTRY,
+                                    CONFDB_MONITOR_ACTIVE_SERVICES,
                                     &ctx->services);
     if (ret != EOK) {
         DEBUG(0, ("No services configured!\n"));
         return EINVAL;
     }
-    ret = append_data_provider(ctx);
-    if (ret != EOK) {
-        DEBUG(0, ("Could not add Data Provider to the list of services!\n"));
-        return ret;
+
+    badsrv = check_services(ctx->services);
+    if (badsrv != NULL) {
+        DEBUG(0, ("Invalid service %s\n", badsrv));
+        return EINVAL;
     }
 
     ctx->domain_ctx = talloc_new(ctx);
     if(!ctx->domain_ctx) {
         return ENOMEM;
     }
-    ret = confdb_get_domains(ctx->cdb, ctx->domain_ctx, &ctx->domains);
+    ret = confdb_get_domains(ctx->cdb, &ctx->domains);
     if (ret != EOK) {
         DEBUG(0, ("No domains configured.\n"));
         return ret;
@@ -914,13 +1022,14 @@ static int get_service_config(struct mt_ctx *ctx, const char *name,
         return ENOMEM;
     }
 
-    path = talloc_asprintf(svc, "config/services/%s", svc->name);
+    path = talloc_asprintf(svc, CONFDB_SERVICE_PATH_TMPL, svc->name);
     if (!path) {
         talloc_free(svc);
         return ENOMEM;
     }
 
-    ret = confdb_get_string(ctx->cdb, svc, path, "command",
+    ret = confdb_get_string(ctx->cdb, svc, path,
+                            CONFDB_SERVICE_COMMAND,
                             NULL, &svc->command);
     if (ret != EOK) {
         DEBUG(0,("Failed to start service '%s'\n", svc->name));
@@ -929,18 +1038,21 @@ static int get_service_config(struct mt_ctx *ctx, const char *name,
     }
 
     if (!svc->command) {
-        svc->command = talloc_asprintf(svc, "%s/sssd_%s -d %d%s",
+        svc->command = talloc_asprintf(svc, "%s/sssd_%s -d %d%s%s",
                                        SSSD_LIBEXEC_PATH,
                                        svc->name, debug_level,
                                        (debug_timestamps?
-                                              " --debug-timestamps":""));
+                                              " --debug-timestamps":""),
+                                       (debug_to_file ?
+                                              " --debug-to-files":""));
         if (!svc->command) {
             talloc_free(svc);
             return ENOMEM;
         }
     }
 
-    ret = confdb_get_int(ctx->cdb, svc, path, "timeout",
+    ret = confdb_get_int(ctx->cdb, svc, path,
+                         CONFDB_SERVICE_TIMEOUT,
                          MONITOR_DEF_PING_TIME, &svc->ping_time);
     if (ret != EOK) {
         DEBUG(0,("Failed to start service '%s'\n", svc->name));
@@ -948,20 +1060,25 @@ static int get_service_config(struct mt_ctx *ctx, const char *name,
         return ret;
     }
 
+    /* 'timeout = 0' should be translated to the default */
+    if (svc->ping_time == 0) {
+        svc->ping_time = MONITOR_DEF_PING_TIME;
+    }
+
     *svc_cfg = svc;
     talloc_free(path);
 
     return EOK;
 }
 
-static int add_new_service(struct mt_ctx *ctx, const char *name, bool startup)
+static int add_new_service(struct mt_ctx *ctx, const char *name)
 {
     int ret;
     struct mt_svc *svc;
 
     ret = get_service_config(ctx, name, &svc);
 
-    ret = start_service(svc, startup);
+    ret = start_service(svc);
     if (ret != EOK) {
         DEBUG(0,("Failed to start service '%s'\n", svc->name));
         talloc_free(svc);
@@ -999,29 +1116,32 @@ static int get_provider_config(struct mt_ctx *ctx, const char *name,
         return ENOMEM;
     }
 
-    path = talloc_asprintf(svc, "config/domains/%s", name);
+    path = talloc_asprintf(svc, CONFDB_DOMAIN_PATH_TMPL, name);
     if (!path) {
         talloc_free(svc);
         return ENOMEM;
     }
 
     ret = confdb_get_string(ctx->cdb, svc, path,
-                            "provider", NULL, &svc->provider);
+                            CONFDB_DOMAIN_ID_PROVIDER,
+                            NULL, &svc->provider);
     if (ret != EOK) {
-        DEBUG(0, ("Failed to find provider from [%s] configuration\n", name));
+        DEBUG(0, ("Failed to find ID provider from [%s] configuration\n", name));
         talloc_free(svc);
         return ret;
     }
 
     ret = confdb_get_string(ctx->cdb, svc, path,
-                            "command", NULL, &svc->command);
+                            CONFDB_DOMAIN_COMMAND,
+                            NULL, &svc->command);
     if (ret != EOK) {
         DEBUG(0, ("Failed to find command from [%s] configuration\n", name));
         talloc_free(svc);
         return ret;
     }
 
-    ret = confdb_get_int(ctx->cdb, svc, path, "timeout",
+    ret = confdb_get_int(ctx->cdb, svc, path,
+                         CONFDB_DOMAIN_TIMEOUT,
                          MONITOR_DEF_PING_TIME, &svc->ping_time);
     if (ret != EOK) {
         DEBUG(0,("Failed to start service '%s'\n", svc->name));
@@ -1029,6 +1149,11 @@ static int get_provider_config(struct mt_ctx *ctx, const char *name,
         return ret;
     }
 
+    /* 'timeout = 0' should be translated to the default */
+    if (svc->ping_time == 0) {
+        svc->ping_time = MONITOR_DEF_PING_TIME;
+    }
+
     talloc_free(path);
 
     /* if no provider is present do not run the domain */
@@ -1040,10 +1165,11 @@ static int get_provider_config(struct mt_ctx *ctx, const char *name,
     /* if there are no custom commands, build a default one */
     if (!svc->command) {
         svc->command = talloc_asprintf(svc,
-                            "%s/sssd_be -d %d%s --provider %s --domain %s",
+                            "%s/sssd_be -d %d%s%s --domain %s",
                             SSSD_LIBEXEC_PATH, debug_level,
                             (debug_timestamps?" --debug-timestamps":""),
-                            svc->provider, svc->name);
+                            (debug_to_file?" --debug-to-files":""),
+                            svc->name);
         if (!svc->command) {
             talloc_free(svc);
             return ENOMEM;
@@ -1054,7 +1180,7 @@ static int get_provider_config(struct mt_ctx *ctx, const char *name,
     return EOK;
 }
 
-static int add_new_provider(struct mt_ctx *ctx, const char *name, bool startup)
+static int add_new_provider(struct mt_ctx *ctx, const char *name)
 {
     int ret;
     struct mt_svc *svc;
@@ -1066,7 +1192,17 @@ static int add_new_provider(struct mt_ctx *ctx, const char *name, bool startup)
         return ret;
     }
 
-    ret = start_service(svc, startup);
+    if (strcasecmp(svc->provider, "local") == 0) {
+        /* The LOCAL provider requires no back-end currently
+         * We'll add it to the service list, but we don't need
+         * to poll it.
+         */
+        svc->svc_started = true;
+        DLIST_ADD(ctx->svc_list, svc);
+        return ENOENT;
+    }
+
+    ret = start_service(svc);
     if (ret != EOK) {
         DEBUG(0,("Failed to start service '%s'\n", svc->name));
         talloc_free(svc);
@@ -1146,7 +1282,7 @@ static int update_monitor_config(struct mt_ctx *ctx)
 
         if (ctx->services[j] == NULL) {
             /* New service added */
-            add_new_service(ctx, new_config->services[i], false);
+            add_new_service(ctx, new_config->services[i]);
         }
         else {
             /* Service already enabled, check for changes */
@@ -1231,7 +1367,7 @@ static int update_monitor_config(struct mt_ctx *ctx)
 
         if (dom == NULL) {
             /* New provider added */
-            add_new_provider(ctx, new_dom->name, false);
+            add_new_provider(ctx, new_dom->name);
         }
         else {
             /* Provider is already in the list.
@@ -1385,6 +1521,86 @@ int read_config_file(const char *config_file)
     return ret;
 }
 
+static errno_t load_configuration(TALLOC_CTX *mem_ctx,
+                                  const char *config_file,
+                                  struct mt_ctx **monitor)
+{
+    errno_t ret;
+    struct mt_ctx *ctx;
+    char *cdb_file = NULL;
+
+    ctx = talloc_zero(mem_ctx, struct mt_ctx);
+    if(!ctx) {
+        return ENOMEM;
+    }
+
+    cdb_file = talloc_asprintf(ctx, "%s/%s", DB_PATH, CONFDB_FILE);
+    if (cdb_file == NULL) {
+        DEBUG(0,("Out of memory, aborting!\n"));
+        ret = ENOMEM;
+        goto done;
+    }
+
+    ret = confdb_init(ctx, &ctx->cdb, cdb_file);
+    if (ret != EOK) {
+        DEBUG(0,("The confdb initialization failed\n"));
+        goto done;
+    }
+
+    /* Initialize the CDB from the configuration file */
+    ret = confdb_test(ctx->cdb);
+    if (ret == ENOENT) {
+        /* First-time setup */
+
+        /* Purge any existing confdb in case an old
+         * misconfiguration gets in the way
+         */
+        talloc_zfree(ctx->cdb);
+        unlink(cdb_file);
+
+        ret = confdb_init(ctx, &ctx->cdb, cdb_file);
+        if (ret != EOK) {
+            DEBUG(0,("The confdb initialization failed\n"));
+            goto done;
+        }
+
+        /* Load special entries */
+        ret = confdb_create_base(ctx->cdb);
+        if (ret != EOK) {
+            DEBUG(0, ("Unable to load special entries into confdb\n"));
+            goto done;
+        }
+    } else if (ret != EOK) {
+        DEBUG(0, ("Fatal error initializing confdb\n"));
+        goto done;
+    }
+    talloc_zfree(cdb_file);
+
+    ret = confdb_init_db(config_file, ctx->cdb);
+    if (ret != EOK) {
+        DEBUG(0, ("ConfDB initialization has failed [%s]\n",
+              strerror(ret)));
+        goto done;
+    }
+
+    /* Validate the configuration in the database */
+    /* Read in the monitor's configuration */
+    ret = get_monitor_config(ctx);
+    if (ret != EOK) {
+        goto done;
+    }
+
+    *monitor = ctx;
+
+    ret = EOK;
+
+done:
+    if (ret != EOK) {
+        talloc_free(ctx);
+    }
+    return ret;
+}
+
 #ifdef HAVE_SYS_INOTIFY_H
 static void process_config_file(struct tevent_context *ev,
                                 struct tevent_timer *te,
@@ -1774,123 +1990,88 @@ static int monitor_config_file(TALLOC_CTX *mem_ctx,
     return EOK;
 }
 
-int monitor_process_init(TALLOC_CTX *mem_ctx,
-                         struct tevent_context *event_ctx,
-                         struct confdb_ctx *cdb,
+int monitor_process_init(struct mt_ctx *ctx,
                          const char *config_file)
 {
-    struct mt_ctx *ctx;
-    struct sysdb_ctx *sysdb;
+    TALLOC_CTX *tmp_ctx;
+    struct sysdb_ctx_list *db_list;
     struct tevent_signal *tes;
-    int ret, i;
-    char *cdb_file;
     struct sss_domain_info *dom;
+    int num_providers;
+    int ret;
 
-    ctx = talloc_zero(mem_ctx, struct mt_ctx);
-    if (!ctx) {
-        DEBUG(0, ("fatal error initializing monitor!\n"));
-        return ENOMEM;
-    }
-    ctx->ev = event_ctx;
-    ctx->cdb = cdb;
+#if 0
+    This feature is incomplete and can leave the SSSD in a bad state if the
+    config file is changed while the SSSD is running.
 
-    /* Initialize the CDB from the configuration file */
-    ret = confdb_test(ctx->cdb);
-    if (ret == ENOENT) {
-        /* First-time setup */
-
-        /* Purge any existing confdb in case an old
-         * misconfiguration gets in the way
-         */
-        talloc_free(ctx->cdb);
-        cdb_file = talloc_asprintf(ctx, "%s/%s", DB_PATH, CONFDB_FILE);
-        if (cdb_file == NULL) {
-            DEBUG(0,("Out of memory, aborting!\n"));
-            talloc_free(ctx);
-            return ENOMEM;
-        }
-
-        unlink(cdb_file);
-        ret = confdb_init(ctx, ctx->ev, &ctx->cdb, cdb_file);
-        if (ret != EOK) {
-            DEBUG(0,("The confdb initialization failed\n"));
-            talloc_free(ctx);
-            return ret;
-        }
-
-        /* Load special entries */
-        ret = confdb_create_base(ctx->cdb);
-        if (ret != EOK) {
-            talloc_free(ctx);
-            return ret;
-        }
-    } else if (ret != EOK) {
-        DEBUG(0, ("Fatal error initializing confdb\n"));
-        talloc_free(ctx);
-        return ret;
-    }
-
-    ret = confdb_init_db(config_file, ctx->cdb);
-    if (ret != EOK) {
-        DEBUG(0, ("ConfDB initialization has failed [%s]\n",
-              strerror(ret)));
-        talloc_free(ctx);
-        return ret;
-    }
-
-    /* Read in the monitor's configuration */
-    ret = get_monitor_config(ctx);
-    if (ret != EOK) {
-        talloc_free(ctx);
-        return ret;
-    }
+    Uncomment this once the backends are honoring reloadConfig()
 
     /* Watch for changes to the confdb config file */
     ret = monitor_config_file(ctx, ctx, config_file, monitor_signal_reconf);
     if (ret != EOK) {
-        talloc_free(ctx);
         return ret;
     }
-
+#endif
     /* Watch for changes to the DNS resolv.conf */
     ret = monitor_config_file(ctx, ctx, RESOLV_CONF_PATH,
                               monitor_update_resolv);
     if (ret != EOK) {
-        talloc_free(ctx);
         return ret;
     }
 
-    /* Avoid a startup race condition between InfoPipe
-     * and NSS. If the sysdb doesn't exist yet, both
-     * will try to create it at the same time. So
-     * we'll have the monitor create it before either of
-     * those processes start.
+    /* Avoid a startup race condition between process.
+     * We need to handle DB upgrades or DB creation only
+     * in one process before all other start.
      */
-    ret = sysdb_init(mem_ctx, ctx->ev, ctx->cdb,
-                     NULL, &sysdb);
+    tmp_ctx = talloc_new(NULL);
+    if (!tmp_ctx) {
+        return ENOMEM;
+    }
+    ret = sysdb_init(tmp_ctx, ctx->ev, ctx->cdb, NULL, true, &db_list);
     if (ret != EOK) {
-        talloc_free(ctx);
         return ret;
     }
-    talloc_free(sysdb);
+    talloc_zfree(tmp_ctx);
 
     /* Initialize D-BUS Server
      * The monitor will act as a D-BUS server for all
      * SSSD processes */
     ret = monitor_dbus_init(ctx);
     if (ret != EOK) {
-        talloc_free(ctx);
         return ret;
     }
 
-    /* then start all services */
-    for (i = 0; ctx->services[i]; i++) {
-        add_new_service(ctx, ctx->services[i], true);
+    /* start providers */
+    num_providers = 0;
+    for (dom = ctx->domains; dom; dom = dom->next) {
+        ret = add_new_provider(ctx, dom->name);
+        if (ret != EOK && ret != ENOENT) {
+            return ret;
+        }
+        if (ret != ENOENT) {
+            num_providers++;
+        }
     }
 
-    /* now start the data providers */
-    for (dom = ctx->domains; dom; dom = dom->next) {
-        add_new_provider(ctx, dom->name, true);
+    if (num_providers > 0) {
+        /* now set the services stratup timeout *
+         * (responders will be started automatically when all
+         *  providers are up and running or when the tomeout
+         *  expires) */
+        ret = add_services_startup_timeout(ctx);
+        if (ret != EOK) {
+            return ret;
+        }
+    } else {
+        int i;
+
+        ctx->services_started = true;
+
+        /* No providers start services immediately
+         * Normally this means only LOCAL is configured */
+        for (i = 0; ctx->services[i]; i++) {
+            add_new_service(ctx, ctx->services[i]);
+        }
     }
 
     /* now start checking for global events */
@@ -1900,7 +2081,6 @@ int monitor_process_init(TALLOC_CTX *mem_ctx,
     tes = tevent_add_signal(ctx->ev, ctx, SIGHUP, 0,
                             monitor_hup, ctx);
     if (tes == NULL) {
-        talloc_free(ctx);
         return EIO;
     }
 
@@ -1908,7 +2088,6 @@ int monitor_process_init(TALLOC_CTX *mem_ctx,
     tes = tevent_add_signal(ctx->ev, ctx, SIGINT, 0,
                             monitor_quit, ctx);
     if (tes == NULL) {
-        talloc_free(ctx);
         return EIO;
     }
 
@@ -1916,7 +2095,6 @@ int monitor_process_init(TALLOC_CTX *mem_ctx,
     tes = tevent_add_signal(ctx->ev, ctx, SIGTERM, 0,
                             monitor_quit, ctx);
     if (tes == NULL) {
-        talloc_free(ctx);
         return EIO;
     }
 
@@ -1963,7 +2141,7 @@ static int monitor_service_init(struct sbus_connection *conn, void *data)
     mini->conn = conn;
 
     /* 5 seconds should be plenty */
-    tv = tevent_timeval_current_ofs(5, 0);
+    tv = tevent_timeval_current_ofs(10, 0);
 
     mini->timeout = tevent_add_timer(ctx->ev, mini, tv, init_timeout, mini);
     if (!mini->timeout) {
@@ -2061,11 +2239,11 @@ static void ping_check(DBusPendingCall *pending, void *data)
     switch (type) {
     case DBUS_MESSAGE_TYPE_METHOD_RETURN:
         /* ok peer replied,
-         * set the reply timestamp into the service structure */
+         * make sure we reset the failure counter in the service structure */
 
         DEBUG(4,("Service %s replied to ping\n", svc->name));
 
-        svc->last_pong = time(NULL);
+        svc->failed_pongs = 0;
         break;
 
     case DBUS_MESSAGE_TYPE_ERROR:
@@ -2232,23 +2410,19 @@ static void service_startup_handler(struct tevent_context *ev,
                                     struct tevent_timer *te,
                                     struct timeval t, void *ptr);
 
-static int start_service(struct mt_svc *svc, bool startup)
+static int start_service(struct mt_svc *svc)
 {
     struct tevent_timer *te;
     struct timeval tv;
 
     DEBUG(4,("Queueing service %s for startup\n", svc->name));
 
-    /* at startup we need to start the data provider service before all others
-     * to avoid races where a service that need dp starts before it is ready
-     * to accept connections on its dbus. So if startup is true delay by 1
-     * second any process that is not the data provider */
+    /* at startup we need to start the data providers before the responders
+     * to avoid races where a service starts before sbus pipes are ready
+     * to accept connections. So if startup is true delay by 2 seconds any
+     * process that is not a data provider */
 
-    if (startup && strcasecmp(svc->name, "dp")) {
-        tv = tevent_timeval_current_ofs(1, 0);
-    } else {
-        tv = tevent_timeval_current();
-    }
+    tv = tevent_timeval_current();
 
     /* Add a timed event to start up the service.
      * We have to do this in order to avoid a race
@@ -2265,13 +2439,6 @@ static int start_service(struct mt_svc *svc, bool startup)
     return EOK;
 }
 
-static int delist_service(void *ptr) {
-    struct mt_svc *svc =
-        talloc_get_type(ptr, struct mt_svc);
-    DLIST_REMOVE(svc->mt_ctx->svc_list, svc);
-    return 0;
-}
-
 static void service_startup_handler(struct tevent_context *ev,
                                     struct tevent_timer *te,
                                     struct timeval t, void *ptr)
@@ -2284,16 +2451,6 @@ static void service_startup_handler(struct tevent_context *ev,
         return;
     }
 
-    if (mt_svc->provider && strcasecmp(mt_svc->provider, "local") == 0) {
-        /* The LOCAL provider requires no back-end currently
-         * We'll add it to the service list, but we don't need
-         * to poll it.
-         */
-        DLIST_ADD(mt_svc->mt_ctx->svc_list, mt_svc);
-        talloc_set_destructor((TALLOC_CTX *)mt_svc, delist_service);
-        return;
-    }
-
     mt_svc->pid = fork();
     if (mt_svc->pid != 0) {
         if (mt_svc->pid == -1) {
@@ -2302,9 +2459,9 @@ static void service_startup_handler(struct tevent_context *ev,
         }
 
         /* Parent */
-        mt_svc->last_pong = time(NULL);
+        mt_svc->mt_ctx->check_children = true;
+        mt_svc->failed_pongs = 0;
         DLIST_ADD(mt_svc->mt_ctx->svc_list, mt_svc);
-        talloc_set_destructor((TALLOC_CTX *)mt_svc, delist_service);
         set_tasks_checker(mt_svc);
 
         return;
@@ -2335,14 +2492,17 @@ int main(int argc, const char *argv[])
     char *config_file = NULL;
     int flags = 0;
     struct main_context *main_ctx;
+    TALLOC_CTX *tmp_ctx;
+    struct mt_ctx *monitor;
     int ret;
+    uid_t uid;
 
     struct poptOption long_options[] = {
         POPT_AUTOHELP
         SSSD_MAIN_OPTS
         {"daemon", 'D', POPT_ARG_NONE, &opt_daemon, 0, \
          "Become a daemon (default)", NULL }, \
-        {"interactive",	'i', POPT_ARG_NONE, &opt_interactive, 0, \
+        {"interactive", 'i', POPT_ARG_NONE, &opt_interactive, 0, \
          "Run interactive (not a daemon)", NULL}, \
         {"config", 'c', POPT_ARG_STRING, &opt_config_file, 0, \
          "Specify a non-default config file", NULL}, \
@@ -2368,33 +2528,62 @@ int main(int argc, const char *argv[])
 
     poptFreeContext(pc);
 
+    uid = getuid();
+    if (uid != 0) {
+        DEBUG(1, ("Running under %d, must be root\n", uid));
+        ERROR("sssd must be run as root\n");
+        return 8;
+    }
+
+    tmp_ctx = talloc_new(NULL);
+    if (!tmp_ctx) {
+        return 7;
+    }
+
     if (opt_daemon) flags |= FLAGS_DAEMON;
     if (opt_interactive) flags |= FLAGS_INTERACTIVE;
 
     if (opt_config_file)
-        config_file = talloc_strdup(NULL, opt_config_file);
+        config_file = talloc_strdup(tmp_ctx, opt_config_file);
     else
-        config_file = talloc_strdup(NULL, CONFDB_DEFAULT_CONFIG_FILE);
+        config_file = talloc_strdup(tmp_ctx, CONFDB_DEFAULT_CONFIG_FILE);
     if(!config_file)
         return 6;
 
     /* we want a pid file check */
     flags |= FLAGS_PID_FILE;
 
+    /* Open before server_setup() does to have logging
+     * during configuration checking */
+    if (debug_to_file) {
+        ret = open_debug_file();
+        if (ret) {
+            return 7;
+        }
+    }
+
     /* Parse config file, fail if cannot be done */
-    ret = read_config_file(config_file);
-    if (ret != EOK) return 4;
+    ret = load_configuration(tmp_ctx, config_file, &monitor);
+    if (ret != EOK) {
+        if (ret == EIO) {
+            DEBUG(1, ("Cannot read configuration file %s\n", config_file));
+            ERROR("Cannot read config file %s, please check if permissions "
+                  "are 0600 and the file is owned by root.root\n", config_file);
+        }
+        return 4;
+    }
 
     /* set up things like debug , signals, daemonization, etc... */
-    ret = server_setup("sssd", flags, MONITOR_CONF_ENTRY, &main_ctx);
+    ret = server_setup("sssd", flags, CONFDB_MONITOR_CONF_ENTRY, &main_ctx);
     if (ret != EOK) return 2;
 
-    ret = monitor_process_init(main_ctx,
-                               main_ctx->event_ctx,
-                               main_ctx->confdb_ctx,
+    monitor->ev = main_ctx->event_ctx;
+    talloc_steal(main_ctx, monitor);
+
+    ret = monitor_process_init(monitor,
                                config_file);
     if (ret != EOK) return 3;
-    talloc_free(config_file);
+    talloc_free(tmp_ctx);
 
     /* loop on main */
     server_loop(main_ctx);
