@@ -30,17 +30,19 @@
 #include "responder/pam/pamsrv.h"
 #include "db/sysdb.h"
 
+static void pam_reply(struct pam_auth_req *preq);
+
 static int extract_authtok(uint32_t *type, uint32_t *size, uint8_t **tok, uint8_t *body, size_t blen, size_t *c) {
-    size_t data_size;
+    uint32_t data_size;
 
     if (blen-(*c) < 2*sizeof(uint32_t)) return EINVAL;
 
-    data_size = ((uint32_t *)&body[*c])[0];
+    memcpy(&data_size, &body[*c], sizeof(uint32_t));
     *c += sizeof(uint32_t);
     if (data_size < sizeof(uint32_t) || (*c)+(data_size) > blen) return EINVAL;
     *size = data_size - sizeof(uint32_t);
 
-    *type = ((uint32_t *)&body[*c])[0];
+    memcpy(type, &body[*c], sizeof(uint32_t));
     *c += sizeof(uint32_t);
 
     *tok = body+(*c);
@@ -56,7 +58,7 @@ static int extract_string(char **var, uint8_t *body, size_t blen, size_t *c) {
 
     if (blen-(*c) < sizeof(uint32_t)+1) return EINVAL;
 
-    size = ((uint32_t *)&body[*c])[0];
+    memcpy(&size, &body[*c], sizeof(uint32_t));
     *c += sizeof(uint32_t);
     if (*c+size > blen) return EINVAL;
 
@@ -71,6 +73,20 @@ static int extract_string(char **var, uint8_t *body, size_t blen, size_t *c) {
     return EOK;
 }
 
+static int extract_uint32_t(uint32_t *var, uint8_t *body, size_t blen, size_t *c) {
+    uint32_t size;
+
+    if (blen-(*c) < 2*sizeof(uint32_t)) return EINVAL;
+
+    memcpy(&size, &body[*c], sizeof(uint32_t));
+    *c += sizeof(uint32_t);
+
+    memcpy(var, &body[*c], sizeof(uint32_t));
+    *c += sizeof(uint32_t);
+
+    return EOK;
+}
+
 static int pam_parse_in_data_v2(struct sss_names_ctx *snctx,
                              struct pam_data *pd,
                              uint8_t *body, size_t blen)
@@ -80,17 +96,18 @@ static int pam_parse_in_data_v2(struct sss_names_ctx *snctx,
     uint32_t size;
     char *pam_user;
     int ret;
+    uint32_t terminator = END_OF_PAM_REQUEST;
 
     if (blen < 4*sizeof(uint32_t)+2 ||
         ((uint32_t *)body)[0] != START_OF_PAM_REQUEST ||
-        ((uint32_t *)(&body[blen - sizeof(uint32_t)]))[0] != END_OF_PAM_REQUEST) {
+        memcmp(&body[blen - sizeof(uint32_t)], &terminator, sizeof(uint32_t)) != 0) {
         DEBUG(1, ("Received data is invalid.\n"));
         return EINVAL;
     }
 
     c = sizeof(uint32_t);
     do {
-        type = ((uint32_t *)&body[c])[0];
+        memcpy(&type, &body[c], sizeof(uint32_t));
         c += sizeof(uint32_t);
         if (c > blen) return EINVAL;
 
@@ -119,6 +136,11 @@ static int pam_parse_in_data_v2(struct sss_names_ctx *snctx,
                 ret = extract_string(&pd->rhost, body, blen, &c);
                 if (ret != EOK) return ret;
                 break;
+            case PAM_ITEM_CLI_PID:
+                ret = extract_uint32_t(&pd->cli_pid,
+                                       body, blen, &c);
+                if (ret != EOK) return ret;
+                break;
             case PAM_ITEM_AUTHTOK:
                 ret = extract_authtok(&pd->authtok_type, &pd->authtok_size,
                                       &pd->authtok, body, blen, &c);
@@ -148,6 +170,26 @@ static int pam_parse_in_data_v2(struct sss_names_ctx *snctx,
 
 }
 
+static int pam_parse_in_data_v3(struct sss_names_ctx *snctx,
+                             struct pam_data *pd,
+                             uint8_t *body, size_t blen)
+{
+    int ret;
+
+    ret = pam_parse_in_data_v2(snctx, pd, body, blen);
+    if (ret != EOK) {
+        DEBUG(1, ("pam_parse_in_data_v2 failed.\n"));
+        return ret;
+    }
+
+    if (pd->cli_pid == 0) {
+        DEBUG(1, ("Missing client PID.\n"));
+        return EINVAL;
+    }
+
+    return EOK;
+}
+
 static int pam_parse_in_data(struct sss_names_ctx *snctx,
                              struct pam_data *pd,
                              uint8_t *body, size_t blen)
@@ -227,7 +269,191 @@ static int pam_parse_in_data(struct sss_names_ctx *snctx,
     return EOK;
 }
 
-static void pam_reply(struct pam_auth_req *preq);
+/*=Save-Last-Login-State===================================================*/
+
+struct set_last_login_state {
+    struct tevent_context *ev;
+    struct sysdb_ctx *dbctx;
+
+    struct sss_domain_info *dom;
+    const char *username;
+    struct sysdb_attrs *attrs;
+
+    struct sysdb_handle *handle;
+
+    struct ldb_result *res;
+};
+
+static void set_last_login_trans_done(struct tevent_req *subreq);
+static void set_last_login_attrs_done(struct tevent_req *subreq);
+static void set_last_login_done(struct tevent_req *subreq);
+
+static struct tevent_req *set_last_login_send(TALLOC_CTX *memctx,
+                                              struct tevent_context *ev,
+                                              struct sysdb_ctx *dbctx,
+                                              struct sss_domain_info *dom,
+                                              const char *username,
+                                              struct sysdb_attrs *attrs)
+{
+    struct tevent_req *req, *subreq;
+    struct set_last_login_state *state;
+
+    req = tevent_req_create(memctx, &state, struct set_last_login_state);
+    if (!req) {
+        return NULL;
+    }
+
+    state->ev = ev;
+    state->dbctx = dbctx;
+    state->dom = dom;
+    state->username = username;
+    state->attrs = attrs;
+    state->handle = NULL;
+
+    subreq = sysdb_transaction_send(state, state->ev, state->dbctx);
+    if (!subreq) {
+        talloc_free(req);
+        return NULL;
+    }
+    tevent_req_set_callback(subreq, set_last_login_trans_done, req);
+
+    return req;
+}
+
+static void set_last_login_trans_done(struct tevent_req *subreq)
+{
+    struct tevent_req *req = tevent_req_callback_data(subreq,
+                                                      struct tevent_req);
+    struct set_last_login_state *state = tevent_req_data(req,
+                                                struct set_last_login_state);
+    int ret;
+
+    ret = sysdb_transaction_recv(subreq, state, &state->handle);
+    talloc_zfree(subreq);
+    if (ret != EOK) {
+        DEBUG(1, ("Unable to acquire sysdb transaction lock\n"));
+        tevent_req_error(req, ret);
+        return;
+    }
+
+    subreq = sysdb_set_user_attr_send(state, state->ev, state->handle,
+                                      state->dom, state->username,
+                                      state->attrs, SYSDB_MOD_REP);
+    if (!subreq) {
+        tevent_req_error(req, ENOMEM);
+        return;
+    }
+    tevent_req_set_callback(subreq, set_last_login_attrs_done, req);
+}
+
+static void set_last_login_attrs_done(struct tevent_req *subreq)
+{
+    struct tevent_req *req = tevent_req_callback_data(subreq,
+                                                      struct tevent_req);
+    struct set_last_login_state *state = tevent_req_data(req,
+                                                struct set_last_login_state);
+    int ret;
+
+    ret = sysdb_set_user_attr_recv(subreq);
+    talloc_zfree(subreq);
+    if (ret != EOK) {
+        DEBUG(4, ("set_user_attr_callback, status [%d][%s]\n",
+                  ret, strerror(ret)));
+        tevent_req_error(req, ret);
+        return;
+    }
+
+    subreq = sysdb_transaction_commit_send(state, state->ev, state->handle);
+    if (!subreq) {
+        tevent_req_error(req, ENOMEM);
+        return;
+    }
+    tevent_req_set_callback(subreq, set_last_login_done, req);
+}
+
+static void set_last_login_done(struct tevent_req *subreq)
+{
+    struct tevent_req *req = tevent_req_callback_data(subreq,
+                                                      struct tevent_req);
+    int ret;
+
+    ret = sysdb_transaction_commit_recv(subreq);
+    if (ret != EOK) {
+        DEBUG(2, ("set_last_login failed.\n"));
+        tevent_req_error(req, ret);
+        return;
+    }
+
+    tevent_req_done(req);
+}
+
+static int set_last_login_recv(struct tevent_req *req)
+{
+    TEVENT_REQ_RETURN_ON_ERROR(req);
+
+    return EOK;
+}
+
+/*=========================================================================*/
+
+
+static void set_last_login_reply(struct tevent_req *req);
+
+static errno_t set_last_login(struct pam_auth_req *preq)
+{
+    struct tevent_req *req;
+    struct sysdb_ctx *dbctx;
+    struct sysdb_attrs *attrs;
+    errno_t ret;
+
+    attrs = sysdb_new_attrs(preq);
+    if (!attrs) {
+        ret = ENOMEM;
+        goto fail;
+    }
+
+    ret = sysdb_attrs_add_time_t(attrs, SYSDB_LAST_ONLINE_AUTH, time(NULL));
+    if (ret != EOK) {
+        goto fail;
+    }
+
+    ret = sysdb_get_ctx_from_list(preq->cctx->rctx->db_list, preq->domain,
+                                  &dbctx);
+    if (ret != EOK) {
+        DEBUG(0, ("Fatal: Sysdb context not found for this domain!\n"));
+        goto fail;
+    }
+
+    req = set_last_login_send(preq, preq->cctx->ev, dbctx,
+                              preq->domain, preq->pd->user, attrs);
+    if (!req) {
+        ret = ENOMEM;
+        goto fail;
+    }
+    tevent_req_set_callback(req, set_last_login_reply, preq);
+
+    return EOK;
+
+fail:
+    return ret;
+}
+
+static void set_last_login_reply(struct tevent_req *req)
+{
+    struct pam_auth_req *preq = tevent_req_callback_data(req,
+                                                         struct pam_auth_req);
+    int ret;
+
+    ret = set_last_login_recv(req);
+    if (ret != EOK) {
+        preq->pd->pam_status = PAM_SYSTEM_ERR;
+    } else {
+        preq->pd->last_auth_saved = true;
+    }
+
+    preq->callback(preq);
+}
+
 static void pam_reply_delay(struct tevent_context *ev, struct tevent_timer *te,
                             struct timeval tv, void *pvt)
 {
@@ -313,6 +539,22 @@ static void pam_reply(struct pam_auth_req *preq)
         return;
     }
 
+    /* If this was a successful login, save the lastLogin time */
+    if (pd->cmd == SSS_PAM_AUTHENTICATE &&
+        pd->pam_status == PAM_SUCCESS &&
+        preq->domain->cache_credentials &&
+        !pd->offline_auth &&
+        !pd->last_auth_saved &&
+        NEED_CHECK_PROVIDER(preq->domain->provider)) {
+        ret = set_last_login(preq);
+        if (ret != EOK) {
+            err = ret;
+            goto done;
+        }
+
+        return;
+    }
+
     ret = sss_packet_new(cctx->creq, 0, sss_packet_get_cmd(cctx->creq->in),
                          &cctx->creq->out);
     if (ret != EOK) {
@@ -382,12 +624,14 @@ static void pam_dom_forwarder(struct pam_auth_req *preq);
 static int pam_forwarder(struct cli_ctx *cctx, int pam_cmd)
 {
     struct sss_domain_info *dom;
+    struct sysdb_ctx *sysdb;
     struct pam_auth_req *preq;
     struct pam_data *pd;
     uint8_t *body;
     size_t blen;
     int timeout;
     int ret;
+    uint32_t terminator = END_OF_PAM_REQUEST;
     preq = talloc_zero(cctx, struct pam_auth_req);
     if (!preq) {
         return ENOMEM;
@@ -403,9 +647,8 @@ static int pam_forwarder(struct cli_ctx *cctx, int pam_cmd)
 
     sss_packet_get_body(cctx->creq->in, &body, &blen);
     if (blen >= sizeof(uint32_t) &&
-        ((uint32_t *)(&body[blen - sizeof(uint32_t)]))[0] != END_OF_PAM_REQUEST) {
+        memcmp(&body[blen - sizeof(uint32_t)], &terminator, sizeof(uint32_t)) != 0) {
         DEBUG(1, ("Received data not terminated.\n"));
-        talloc_free(preq);
         ret = EINVAL;
         goto done;
     }
@@ -420,13 +663,15 @@ static int pam_forwarder(struct cli_ctx *cctx, int pam_cmd)
         case 2:
             ret = pam_parse_in_data_v2(cctx->rctx->names, pd, body, blen);
             break;
+        case 3:
+            ret = pam_parse_in_data_v3(cctx->rctx->names, pd, body, blen);
+            break;
         default:
             DEBUG(1, ("Illegal protocol version [%d].\n",
                       cctx->cli_protocol_version->version));
             ret = EINVAL;
     }
     if (ret != EOK) {
-        talloc_free(preq);
         ret = EINVAL;
         goto done;
     }
@@ -437,7 +682,6 @@ static int pam_forwarder(struct cli_ctx *cctx, int pam_cmd)
             if (strcasecmp(dom->name, pd->domain) == 0) break;
         }
         if (!dom) {
-            talloc_free(preq);
             ret = ENOENT;
             goto done;
         }
@@ -480,13 +724,20 @@ static int pam_forwarder(struct cli_ctx *cctx, int pam_cmd)
 
         ret = sss_dp_send_acct_req(preq->cctx->rctx, preq,
                                    pam_check_user_dp_callback, preq,
-                                   timeout, preq->domain->name, SSS_DP_USER,
+                                   timeout, preq->domain->name,
+                                   false, SSS_DP_INITGROUPS,
                                    preq->pd->user, 0);
     }
     else {
         preq->check_provider = NEED_CHECK_PROVIDER(preq->domain->provider);
 
-        ret = sysdb_getpwnam(preq, cctx->rctx->sysdb,
+        ret = sysdb_get_ctx_from_list(cctx->rctx->db_list,
+                                      preq->domain, &sysdb);
+        if (ret != EOK) {
+            DEBUG(0, ("Fatal: Sysdb CTX not found for this domain!\n"));
+            goto done;
+        }
+        ret = sysdb_getpwnam(preq, sysdb,
                              preq->domain, preq->pd->user,
                              pam_check_user_callback, preq);
     }
@@ -508,30 +759,24 @@ static void pam_check_user_dp_callback(uint16_t err_maj, uint32_t err_min,
                                        const char *err_msg, void *ptr)
 {
     struct pam_auth_req *preq = talloc_get_type(ptr, struct pam_auth_req);
-    struct ldb_result *res = NULL;
+    struct sysdb_ctx *sysdb;
     int ret;
 
-    if ((err_maj != DP_ERR_OK) && (err_maj != DP_ERR_OFFLINE)) {
+    if (err_maj) {
         DEBUG(2, ("Unable to get information from Data Provider\n"
                   "Error: %u, %u, %s\n",
                   (unsigned int)err_maj, (unsigned int)err_min, err_msg));
-        ret = EFAULT;
+    }
+
+    /* always try to see if we have the user in cache even if the provider
+     * returned an error */
+    ret = sysdb_get_ctx_from_list(preq->cctx->rctx->db_list,
+                                  preq->domain, &sysdb);
+    if (ret != EOK) {
+        DEBUG(0, ("Fatal: Sysdb CTX not found for this domain!\n"));
         goto done;
     }
-
-    if (err_maj == DP_ERR_OFFLINE) {
-         if (preq->data) res = talloc_get_type(preq->data, struct ldb_result);
-         if (!res) res = talloc_zero(preq, struct ldb_result);
-         if (!res) {
-            ret = EFAULT;
-            goto done;
-        }
-
-        pam_check_user_callback(preq, LDB_SUCCESS, res);
-        return;
-    }
-
-    ret = sysdb_getpwnam(preq, preq->cctx->rctx->sysdb,
+    ret = sysdb_getpwnam(preq, sysdb,
                          preq->domain, preq->pd->user,
                          pam_check_user_callback, preq);
 
@@ -547,10 +792,10 @@ static void pam_check_user_callback(void *ptr, int status,
 {
     struct pam_auth_req *preq = talloc_get_type(ptr, struct pam_auth_req);
     struct sss_domain_info *dom;
-    uint64_t lastUpdate;
+    struct sysdb_ctx *sysdb;
+    uint64_t cacheExpire;
     bool call_provider = false;
     time_t timeout;
-    time_t cache_timeout;
     int ret;
 
     if (status != LDB_SUCCESS) {
@@ -568,11 +813,9 @@ static void pam_check_user_callback(void *ptr, int status,
             break;
 
         case 1:
-            cache_timeout = 30; /* FIXME: read from conf */
-
-            lastUpdate = ldb_msg_find_attr_as_uint64(res->msgs[0],
-                                                     SYSDB_LAST_UPDATE, 0);
-            if (lastUpdate + cache_timeout < time(NULL)) {
+            cacheExpire = ldb_msg_find_attr_as_uint64(res->msgs[0],
+                                                      SYSDB_CACHE_EXPIRE, 0);
+            if (cacheExpire < time(NULL)) {
                 call_provider = true;
             }
             break;
@@ -597,7 +840,8 @@ static void pam_check_user_callback(void *ptr, int status,
 
         ret = sss_dp_send_acct_req(preq->cctx->rctx, preq,
                                    pam_check_user_dp_callback, preq,
-                                   timeout, preq->domain->name, SSS_DP_USER,
+                                   timeout, preq->domain->name,
+                                   false, SSS_DP_USER,
                                    preq->pd->user, 0);
         if (ret != EOK) {
             DEBUG(3, ("Failed to dispatch request: %d(%s)\n",
@@ -667,13 +911,21 @@ static void pam_check_user_callback(void *ptr, int status,
                                                pam_check_user_dp_callback,
                                                preq, timeout,
                                                preq->domain->name,
-                                               SSS_DP_USER,
+                                               false, SSS_DP_USER,
                                                preq->pd->user, 0);
                 }
                 else {
                     preq->check_provider = NEED_CHECK_PROVIDER(preq->domain->provider);
 
-                    ret = sysdb_getpwnam(preq, preq->cctx->rctx->sysdb,
+                    ret = sysdb_get_ctx_from_list(preq->cctx->rctx->db_list,
+                                                  preq->domain, &sysdb);
+                    if (ret != EOK) {
+                        DEBUG(0, ("Fatal: Sysdb CTX not found for this domain!\n"));
+                        preq->pd->pam_status = PAM_SYSTEM_ERR;
+                        pam_reply(preq);
+                        return;
+                    }
+                    ret = sysdb_getpwnam(preq, sysdb,
                                          preq->domain, preq->pd->user,
                                          pam_check_user_callback, preq);
                 }
@@ -798,11 +1050,17 @@ static int pam_cmd_chauthtok(struct cli_ctx *cctx) {
     return pam_forwarder(cctx, SSS_PAM_CHAUTHTOK);
 }
 
+static int pam_cmd_chauthtok_prelim(struct cli_ctx *cctx) {
+    DEBUG(4, ("entering pam_cmd_chauthtok_prelim\n"));
+    return pam_forwarder(cctx, SSS_PAM_CHAUTHTOK_PRELIM);
+}
+
 struct cli_protocol_version *register_cli_protocol_version(void)
 {
     static struct cli_protocol_version pam_cli_protocol_version[] = {
-        {1, "2008-09-05", "initial version, \\0 terminated strings"},
+        {3, "2009-09-14", "make cli_pid mandatory"},
         {2, "2009-05-12", "new format <type><size><data>"},
+        {1, "2008-09-05", "initial version, \\0 terminated strings"},
         {0, NULL, NULL}
     };
 
@@ -819,6 +1077,7 @@ struct sss_cmd_table *get_pam_cmds(void)
         {SSS_PAM_OPEN_SESSION, pam_cmd_open_session},
         {SSS_PAM_CLOSE_SESSION, pam_cmd_close_session},
         {SSS_PAM_CHAUTHTOK, pam_cmd_chauthtok},
+        {SSS_PAM_CHAUTHTOK_PRELIM, pam_cmd_chauthtok_prelim},
         {SSS_CLI_NULL, NULL}
     };
 
