@@ -1,3 +1,23 @@
+/*
+    Authors:
+        Sumit Bose <sbose@redhat.com>
+
+    Copyright (C) 2009 Red Hat
+
+    This program is free software; you can redistribute it and/or modify
+    it under the terms of the GNU Lesser General Public License as published by
+    the Free Software Foundation; either version 3 of the License, or
+    (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public License
+    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+*/
+
 #define PAM_SM_AUTH
 #define PAM_SM_ACCOUNT
 #define PAM_SM_SESSION
@@ -8,11 +28,10 @@
 #include <stdlib.h>
 #include <stdint.h>
 #include <syslog.h>
-#include <locale.h>
 
 #include <security/pam_modules.h>
-#include <security/pam_misc.h>
 #include <security/pam_ext.h>
+#include <security/pam_modutil.h>
 #include "sss_pam_macros.h"
 
 #include "sss_cli.h"
@@ -44,8 +63,8 @@ struct pam_items {
     size_t pam_authtok_size;
     int pam_newauthtok_type;
     size_t pam_newauthtok_size;
-    char *pam_cli_locale;
-    size_t pam_cli_locale_size;
+    pid_t cli_pid;
+    const char *login_name;
 };
 
 #define DEBUG_MGS_LEN 1024
@@ -86,16 +105,20 @@ static size_t add_authtok_item(enum pam_item_type type,
                                const char *tok, const size_t size,
                                uint8_t *buf) {
     size_t rp=0;
+    uint32_t c;
 
     if (tok == NULL) return 0;
 
-    ((uint32_t *)(&buf[rp]))[0] = type;
+    c = type;
+    memcpy(&buf[rp], &c, sizeof(uint32_t));
     rp += sizeof(uint32_t);
 
-    ((uint32_t *)(&buf[rp]))[0] = size + sizeof(uint32_t);
+    c = size + sizeof(uint32_t);
+    memcpy(&buf[rp], &c, sizeof(uint32_t));
     rp += sizeof(uint32_t);
 
-    ((uint32_t *)(&buf[rp]))[0] = authtok_type;
+    c = authtok_type;
+    memcpy(&buf[rp], &c, sizeof(uint32_t));
     rp += sizeof(uint32_t);
 
     memcpy(&buf[rp], tok, size);
@@ -104,16 +127,40 @@ static size_t add_authtok_item(enum pam_item_type type,
     return rp;
 }
 
+
+static size_t add_uint32_t_item(enum pam_item_type type, const uint32_t val,
+                                uint8_t *buf) {
+    size_t rp=0;
+    uint32_t c;
+
+    c = type;
+    memcpy(&buf[rp], &c, sizeof(uint32_t));
+    rp += sizeof(uint32_t);
+
+    c = sizeof(uint32_t);
+    memcpy(&buf[rp], &c, sizeof(uint32_t));
+    rp += sizeof(uint32_t);
+
+    c = val;
+    memcpy(&buf[rp], &c, sizeof(uint32_t));
+    rp += sizeof(uint32_t);
+
+    return rp;
+}
+
 static size_t add_string_item(enum pam_item_type type, const char *str,
                            const size_t size, uint8_t *buf) {
     size_t rp=0;
+    uint32_t c;
 
-    if (*str == '\0') return 0;
+    if (str == NULL || *str == '\0') return 0;
 
-    ((uint32_t *)(&buf[rp]))[0] = type;
+    c = type;
+    memcpy(&buf[rp], &c, sizeof(uint32_t));
     rp += sizeof(uint32_t);
 
-    ((uint32_t *)(&buf[rp]))[0] = size;
+    c = size;
+    memcpy(&buf[rp], &c, sizeof(uint32_t));
     rp += sizeof(uint32_t);
 
     memcpy(&buf[rp], str, size);
@@ -122,29 +169,44 @@ static size_t add_string_item(enum pam_item_type type, const char *str,
     return rp;
 }
 
-static int pack_message_v2(struct pam_items *pi, size_t *size,
+static void overwrite_and_free_authtoks(struct pam_items *pi)
+{
+    if (pi->pam_authtok != NULL) {
+        _pam_overwrite_n((void *)pi->pam_authtok, pi->pam_authtok_size);
+        free((void *)pi->pam_authtok);
+        pi->pam_authtok = NULL;
+    }
+
+    if (pi->pam_newauthtok != NULL) {
+        _pam_overwrite_n((void *)pi->pam_newauthtok,  pi->pam_newauthtok_size);
+        free((void *)pi->pam_newauthtok);
+        pi->pam_newauthtok = NULL;
+    }
+}
+
+static int pack_message_v3(struct pam_items *pi, size_t *size,
                            uint8_t **buffer) {
     int len;
     uint8_t *buf;
     int rp;
+    uint32_t terminator = END_OF_PAM_REQUEST;
 
     len = sizeof(uint32_t) +
           2*sizeof(uint32_t) + pi->pam_user_size +
           sizeof(uint32_t);
-    len +=  *pi->pam_service != '\0' ?
+    len += *pi->pam_service != '\0' ?
                 2*sizeof(uint32_t) + pi->pam_service_size : 0;
-    len +=  *pi->pam_tty != '\0' ?
+    len += *pi->pam_tty != '\0' ?
                 2*sizeof(uint32_t) + pi->pam_tty_size : 0;
-    len +=  *pi->pam_ruser != '\0' ?
+    len += *pi->pam_ruser != '\0' ?
                 2*sizeof(uint32_t) + pi->pam_ruser_size : 0;
-    len +=  *pi->pam_rhost != '\0' ?
+    len += *pi->pam_rhost != '\0' ?
                 2*sizeof(uint32_t) + pi->pam_rhost_size : 0;
-    len +=  *pi->pam_cli_locale != '\0' ?
-                2*sizeof(uint32_t) + pi->pam_cli_locale_size : 0;
-    len +=  pi->pam_authtok != NULL ?
+    len += pi->pam_authtok != NULL ?
                 3*sizeof(uint32_t) + pi->pam_authtok_size : 0;
-    len +=  pi->pam_newauthtok != NULL ?
+    len += pi->pam_newauthtok != NULL ?
                 3*sizeof(uint32_t) + pi->pam_newauthtok_size : 0;
+    len += 3*sizeof(uint32_t); /* cli_pid */
 
     buf = malloc(len);
     if (buf == NULL) {
@@ -171,23 +233,16 @@ static int pack_message_v2(struct pam_items *pi, size_t *size,
     rp += add_string_item(PAM_ITEM_RHOST, pi->pam_rhost, pi->pam_rhost_size,
                           &buf[rp]);
 
-    rp += add_string_item(PAM_CLI_LOCALE, pi->pam_cli_locale,
-                          pi->pam_cli_locale_size, &buf[rp]);
+    rp += add_uint32_t_item(PAM_ITEM_CLI_PID, (uint32_t) pi->cli_pid, &buf[rp]);
 
     rp += add_authtok_item(PAM_ITEM_AUTHTOK, pi->pam_authtok_type,
                            pi->pam_authtok, pi->pam_authtok_size, &buf[rp]);
-    _pam_overwrite_n((void *)pi->pam_authtok, pi->pam_authtok_size);
-    free((void *)pi->pam_authtok);
-    pi->pam_authtok = NULL;
 
     rp += add_authtok_item(PAM_ITEM_NEWAUTHTOK, pi->pam_newauthtok_type,
                            pi->pam_newauthtok, pi->pam_newauthtok_size,
                            &buf[rp]);
-    _pam_overwrite_n((void *)pi->pam_newauthtok,  pi->pam_newauthtok_size);
-    free((void *)pi->pam_newauthtok);
-    pi->pam_newauthtok = NULL;
 
-    ((uint32_t *)(&buf[rp]))[0] = END_OF_PAM_REQUEST;
+    memcpy(&buf[rp], &terminator, sizeof(uint32_t));
     rp += sizeof(uint32_t);
 
     if (rp != len) {
@@ -318,52 +373,47 @@ static int eval_response(pam_handle_t *pamh, size_t buflen, uint8_t *buf)
     int ret;
     size_t p=0;
     char *env_item;
-    int32_t *c;
-    int32_t *type;
-    int32_t *len;
-    int32_t *pam_status;
+    int32_t c;
+    int32_t type;
+    int32_t len;
+    int32_t pam_status;
 
     if (buflen < (2*sizeof(int32_t))) {
         D(("response buffer is too small"));
         return PAM_BUF_ERR;
     }
 
-    pam_status = ((int32_t *)(buf+p));
+    memcpy(&pam_status, buf+p, sizeof(int32_t));
     p += sizeof(int32_t);
 
 
-    c = ((int32_t *)(buf+p));
+    memcpy(&c, buf+p, sizeof(int32_t));
     p += sizeof(int32_t);
 
-    while(*c>0) {
+    while(c>0) {
         if (buflen < (p+2*sizeof(int32_t))) {
             D(("response buffer is too small"));
             return PAM_BUF_ERR;
         }
 
-        type = ((int32_t *)(buf+p));
+        memcpy(&type, buf+p, sizeof(int32_t));
         p += sizeof(int32_t);
 
-        len = ((int32_t *)(buf+p));
+        memcpy(&len, buf+p, sizeof(int32_t));
         p += sizeof(int32_t);
 
-        if (buflen < (p + *len)) {
+        if (buflen < (p + len)) {
             D(("response buffer is too small"));
             return PAM_BUF_ERR;
         }
 
-        switch(*type) {
+        switch(type) {
             case PAM_USER_INFO:
-                if (buf[p + (*len -1)] != '\0') {
+                if (buf[p + (len -1)] != '\0') {
                     D(("user info does not end with \\0."));
                     break;
                 }
                 logger(pamh, LOG_INFO, "user info: [%s]", &buf[p]);
-                ret = do_pam_conversation(pamh, PAM_TEXT_INFO, (char *) &buf[p],
-                                          NULL, NULL);
-                if (ret != PAM_SUCCESS) {
-                    D(("do_pam_conversation, canot display user info."));
-                }
                 break;
             case PAM_DOMAIN_NAME:
                 D(("domain name: [%s]", &buf[p]));
@@ -371,13 +421,13 @@ static int eval_response(pam_handle_t *pamh, size_t buflen, uint8_t *buf)
             case ENV_ITEM:
             case PAM_ENV_ITEM:
             case ALL_ENV_ITEM:
-                if (buf[p + (*len -1)] != '\0') {
+                if (buf[p + (len -1)] != '\0') {
                     D(("env item does not end with \\0."));
                     break;
                 }
 
                 D(("env item: [%s]", &buf[p]));
-                if (*type == PAM_ENV_ITEM || *type == ALL_ENV_ITEM) {
+                if (type == PAM_ENV_ITEM || type == ALL_ENV_ITEM) {
                     ret = pam_putenv(pamh, (char *)&buf[p]);
                     if (ret != PAM_SUCCESS) {
                         D(("pam_putenv failed."));
@@ -385,7 +435,7 @@ static int eval_response(pam_handle_t *pamh, size_t buflen, uint8_t *buf)
                     }
                 }
 
-                if (*type == ENV_ITEM || *type == ALL_ENV_ITEM) {
+                if (type == ENV_ITEM || type == ALL_ENV_ITEM) {
                     env_item = strdup((char *)&buf[p]);
                     if (env_item == NULL) {
                         D(("strdup failed"));
@@ -399,9 +449,9 @@ static int eval_response(pam_handle_t *pamh, size_t buflen, uint8_t *buf)
                 }
                 break;
         }
-        p += *len;
+        p += len;
 
-        --(*c);
+        --c;
     }
 
     return PAM_SUCCESS;
@@ -410,7 +460,6 @@ static int eval_response(pam_handle_t *pamh, size_t buflen, uint8_t *buf)
 static int get_pam_items(pam_handle_t *pamh, struct pam_items *pi)
 {
     int ret;
-    char *cli_locale;
 
     pi->pam_authtok_type = SSS_AUTHTOK_TYPE_EMPTY;
     pi->pam_authtok = NULL;
@@ -462,14 +511,10 @@ static int get_pam_items(pam_handle_t *pamh, struct pam_items *pi)
     if (ret != PAM_SUCCESS) return ret;
     if (pi->pamstack_oldauthtok == NULL) pi->pamstack_oldauthtok="";
 
-    cli_locale = setlocale(LC_ALL, NULL);
-    if (cli_locale == NULL) {
-        pi->pam_cli_locale = strdup("");
-    } else {
-        pi->pam_cli_locale = strdup(cli_locale);
-        if (pi->pam_cli_locale == NULL) return PAM_BUF_ERR;
-    }
-    pi->pam_cli_locale_size = strlen(pi->pam_cli_locale)+1;
+    pi->cli_pid = getpid();
+
+    pi->login_name = pam_modutil_getlogin(pamh);
+    if (pi->login_name == NULL) pi->login_name="";
 
     return PAM_SUCCESS;
 }
@@ -489,7 +534,7 @@ static void print_pam_items(struct pam_items *pi)
             CHECK_AND_RETURN_PI_STRING(pi->pamstack_oldauthtok)));
     D(("Authtok: %s", CHECK_AND_RETURN_PI_STRING(pi->pam_authtok)));
     D(("Newauthtok: %s", CHECK_AND_RETURN_PI_STRING(pi->pam_newauthtok)));
-    D(("Locale: %s", CHECK_AND_RETURN_PI_STRING(pi->pam_cli_locale)));
+    D(("Cli_PID: %d", pi->cli_pid));
 }
 
 static int send_and_receive(pam_handle_t *pamh, struct pam_items *pi,
@@ -505,7 +550,7 @@ static int send_and_receive(pam_handle_t *pamh, struct pam_items *pi,
 
     print_pam_items(pi);
 
-    ret = pack_message_v2(pi, &rd.len, &buf);
+    ret = pack_message_v3(pi, &rd.len, &buf);
     if (ret != 0) {
         D(("pack_message failed."));
         pam_status = PAM_SYSTEM_ERR;
@@ -536,8 +581,15 @@ static int send_and_receive(pam_handle_t *pamh, struct pam_items *pi,
         goto done;
     }
     logger(pamh, (pam_status == PAM_SUCCESS ? LOG_INFO : LOG_NOTICE),
-           "received for user %s: %d (%s)", pi->pam_user, pam_status,
-           pam_strerror(pamh,pam_status));
+           "authentication %s; logname=%s uid=%d euid=%d tty=%s ruser=%s "
+           "rhost=%s user=%s",
+           pam_status == PAM_SUCCESS ? "success" : "failure",
+           pi->login_name, getuid(), geteuid(), pi->pam_tty, pi->pam_ruser,
+           pi->pam_rhost, pi->pam_user);
+    if (pam_status != PAM_SUCCESS) {
+           logger(pamh, LOG_NOTICE, "received for user %s: %d (%s)",
+                  pi->pam_user, pam_status, pam_strerror(pamh,pam_status));
+    }
 
 done:
     if (buf != NULL ) {
@@ -630,6 +682,112 @@ static void eval_argv(pam_handle_t *pamh, int argc, const char **argv,
     return;
 }
 
+static int get_authtok_for_authentication(pam_handle_t *pamh,
+                                          struct pam_items *pi,
+                                          uint32_t flags)
+{
+    int ret;
+
+    if (flags & FLAGS_USE_FIRST_PASS) {
+        pi->pam_authtok_type = SSS_AUTHTOK_TYPE_PASSWORD;
+        pi->pam_authtok = strdup(pi->pamstack_authtok);
+        if (pi->pam_authtok == NULL) {
+            D(("option use_first_pass set, but no password found"));
+            return PAM_BUF_ERR;
+        }
+        pi->pam_authtok_size = strlen(pi->pam_authtok);
+    } else {
+        ret = prompt_password(pamh, pi);
+        if (ret != PAM_SUCCESS) {
+            D(("failed to get password from user"));
+            return ret;
+        }
+
+        if (flags & FLAGS_FORWARD_PASS) {
+            ret = pam_set_item(pamh, PAM_AUTHTOK, pi->pam_authtok);
+            if (ret != PAM_SUCCESS) {
+                D(("Failed to set PAM_AUTHTOK [%s], "
+                   "authtok may not be available for other modules",
+                   pam_strerror(pamh,ret)));
+            }
+        }
+    }
+
+    return PAM_SUCCESS;
+}
+
+static int get_authtok_for_password_change(pam_handle_t *pamh,
+                                           struct pam_items *pi,
+                                           uint32_t flags,
+                                           int pam_flags)
+{
+    int ret;
+
+    /* we query for the old password during PAM_PRELIM_CHECK to make
+     * pam_sss work e.g. with pam_cracklib */
+    if (pam_flags & PAM_PRELIM_CHECK) {
+        if (getuid() != 0 && !(flags & FLAGS_USE_FIRST_PASS)) {
+            ret = prompt_password(pamh, pi);
+            if (ret != PAM_SUCCESS) {
+                D(("failed to get password from user"));
+                return ret;
+            }
+
+            ret = pam_set_item(pamh, PAM_OLDAUTHTOK, pi->pam_authtok);
+            if (ret != PAM_SUCCESS) {
+                D(("Failed to set PAM_OLDAUTHTOK [%s], "
+                   "oldauthtok may not be available",
+                   pam_strerror(pamh,ret)));
+                   return ret;
+            }
+        }
+
+        return PAM_SUCCESS;
+    }
+
+    if (pi->pamstack_oldauthtok == NULL) {
+        if (getuid() != 0) {
+            D(("no password found for chauthtok"));
+            return PAM_BUF_ERR;
+        } else {
+            pi->pam_authtok_type = SSS_AUTHTOK_TYPE_EMPTY;
+            pi->pam_authtok = NULL;
+            pi->pam_authtok_size = 0;
+        }
+    } else {
+        pi->pam_authtok = strdup(pi->pamstack_oldauthtok);
+        pi->pam_authtok_type = SSS_AUTHTOK_TYPE_PASSWORD;
+        pi->pam_authtok_size = strlen(pi->pam_authtok);
+    }
+
+    if (flags & FLAGS_USE_AUTHTOK) {
+        pi->pam_newauthtok_type = SSS_AUTHTOK_TYPE_PASSWORD;
+        pi->pam_newauthtok =  strdup(pi->pamstack_authtok);
+        if (pi->pam_newauthtok == NULL) {
+            D(("option use_authtok set, but no new password found"));
+            return PAM_BUF_ERR;
+        }
+        pi->pam_newauthtok_size = strlen(pi->pam_newauthtok);
+    } else {
+        ret = prompt_new_password(pamh, pi);
+        if (ret != PAM_SUCCESS) {
+            D(("failed to get new password from user"));
+            return ret;
+        }
+
+        if (flags & FLAGS_FORWARD_PASS) {
+            ret = pam_set_item(pamh, PAM_AUTHTOK, pi->pam_newauthtok);
+            if (ret != PAM_SUCCESS) {
+                D(("Failed to set PAM_AUTHTOK [%s], "
+                   "oldauthtok may not be available",
+                   pam_strerror(pamh,ret)));
+            }
+        }
+    }
+
+    return PAM_SUCCESS;
+}
+
 static int pam_sss(enum sss_cli_command task, pam_handle_t *pamh,
                    int pam_flags, int argc, const char **argv)
 {
@@ -652,82 +810,23 @@ static int pam_sss(enum sss_cli_command task, pam_handle_t *pamh,
 
     switch(task) {
         case SSS_PAM_AUTHENTICATE:
-            if (flags & FLAGS_USE_FIRST_PASS) {
-                pi.pam_authtok_type = SSS_AUTHTOK_TYPE_PASSWORD;
-                pi.pam_authtok = strdup(pi.pamstack_authtok);
-                if (pi.pam_authtok == NULL) {
-                    D(("option use_first_pass set, but no password found"));
-                    return PAM_BUF_ERR;
-                }
-                pi.pam_authtok_size = strlen(pi.pam_authtok);
-            } else {
-                prompt_password(pamh, &pi);
-
-                if (flags & FLAGS_FORWARD_PASS) {
-                    ret = pam_set_item(pamh, PAM_AUTHTOK, pi.pam_authtok);
-                    if (ret != PAM_SUCCESS) {
-                        D(("Failed to set PAM_AUTHTOK [%s], "
-                           "authtok may not be available for other modules",
-                           pam_strerror(pamh,ret)));
-                    }
-                }
+            ret = get_authtok_for_authentication(pamh, &pi, flags);
+            if (ret != PAM_SUCCESS) {
+                D(("failed to get authentication token: %s",
+                   pam_strerror(pamh, ret)));
+                return ret;
             }
-
             break;
         case SSS_PAM_CHAUTHTOK:
-            /* we query for the old password during PAM_PRELIM_CHECK to make
-             * pam_sss work e.g. with pam_cracklib */
+            ret = get_authtok_for_password_change(pamh, &pi, flags, pam_flags);
+            if (ret != PAM_SUCCESS) {
+                D(("failed to get tokens for password change: %s",
+                   pam_strerror(pamh, ret)));
+                return ret;
+            }
             if (pam_flags & PAM_PRELIM_CHECK) {
-                if (getuid() != 0 && !(flags & FLAGS_USE_FIRST_PASS)) {
-                    prompt_password(pamh, &pi);
-
-                    ret = pam_set_item(pamh, PAM_OLDAUTHTOK, pi.pam_authtok);
-                    if (ret != PAM_SUCCESS) {
-                        D(("Failed to set PAM_OLDAUTHTOK [%s], "
-                           "oldauthtok may not be available",
-                           pam_strerror(pamh,ret)));
-                           return ret;
-                    }
-                }
-
-                return PAM_SUCCESS;
+                task = SSS_PAM_CHAUTHTOK_PRELIM;
             }
-
-            if (getuid() != 0) {
-                pi.pam_authtok_type = SSS_AUTHTOK_TYPE_PASSWORD;
-                pi.pam_authtok = strdup(pi.pamstack_oldauthtok);
-                if (pi.pam_authtok == NULL) {
-                    D(("no password found for chauthtok"));
-                    return PAM_BUF_ERR;
-                }
-                pi.pam_authtok_size = strlen(pi.pam_authtok);
-            } else {
-                pi.pam_authtok_type = SSS_AUTHTOK_TYPE_EMPTY;
-                pi.pam_authtok = NULL;
-                pi.pam_authtok_size = 0;
-            }
-
-            if (flags & FLAGS_USE_AUTHTOK) {
-                pi.pam_newauthtok_type = SSS_AUTHTOK_TYPE_PASSWORD;
-                pi.pam_newauthtok =  strdup(pi.pamstack_authtok);
-                if (pi.pam_newauthtok == NULL) {
-                    D(("option use_authtok set, but no new password found"));
-                    return PAM_BUF_ERR;
-                }
-                pi.pam_newauthtok_size = strlen(pi.pam_newauthtok);
-            } else {
-                prompt_new_password(pamh, &pi);
-
-                if (flags & FLAGS_FORWARD_PASS) {
-                    ret = pam_set_item(pamh, PAM_AUTHTOK, pi.pam_newauthtok);
-                    if (ret != PAM_SUCCESS) {
-                        D(("Failed to set PAM_AUTHTOK [%s], "
-                           "oldauthtok may not be available",
-                           pam_strerror(pamh,ret)));
-                    }
-                }
-            }
-
             break;
         case SSS_PAM_ACCT_MGMT:
         case SSS_PAM_SETCRED:
@@ -739,7 +838,31 @@ static int pam_sss(enum sss_cli_command task, pam_handle_t *pamh,
             return PAM_SYSTEM_ERR;
     }
 
-    return send_and_receive(pamh, &pi, task);
+    ret = send_and_receive(pamh, &pi, task);
+
+    if (ret == PAM_AUTHTOK_EXPIRED && task == SSS_PAM_AUTHENTICATE) {
+        D(("Authtoken expired, trying to change it"));
+        ret = do_pam_conversation(pamh, PAM_ERROR_MSG,
+                                  _("Password has expired."), NULL, NULL);
+        if (ret != PAM_SUCCESS) {
+            D(("do_pam_conversation failed."));
+            return PAM_SYSTEM_ERR;
+        }
+
+        pi.pamstack_oldauthtok = pi.pam_authtok;
+        ret = get_authtok_for_password_change(pamh, &pi, flags, pam_flags);
+        if (ret != PAM_SUCCESS) {
+            D(("failed to get tokens for password change: %s",
+               pam_strerror(pamh, ret)));
+            return ret;
+        }
+
+        ret = send_and_receive(pamh, &pi, SSS_PAM_CHAUTHTOK);
+    }
+
+    overwrite_and_free_authtoks(&pi);
+
+    return ret;
 }
 
 PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
